#
# Copyright 2019 Universidad Complutense de Madrid
#
# This file is part of EMIR DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from numina.instrument.hwdevice import HWDevice
from numina.array.bbox import BoundingBox
from numina.instrument.generic import ElementBase


class BarsModel(ElementBase):
    def __init__(self, name):
        super(BarsModel, self).__init__(name)


class ConfigurableSlitUnit(HWDevice):

    OPEN_L_LIMIT = 6
    OPEN_R_LIMIT = 2057
    NBARS = 55

    def __init__(self, name, barsmodel, parent=None):
        super(ConfigurableSlitUnit, self).__init__(name, parent=parent)

        # Attributtes added to read from JSON
        self.origin = None
        self.configurations = {}
        # Up to HERE

        self.lbars = []
        self.rbars = []
        self.slits = {}

        nominal = barsmodel.values['nominal']

        for i in range(self.NBARS):
            name = "BarL_{}".format(i + 1)
            _, ypos, delt, x0 = nominal[i]
            model = pol.Polynomial([x0, delt])
            bar = CSUBarL(name, ypos, model, parent=self)
            self.lbars.append(bar)

        for i in range(self.NBARS):
            name = "BarR_{}".format(self.NBARS + i + 1)
            _, ypos, delt, x0 = nominal[i]
            model = pol.Polynomial([x0, delt])
            bar = CSUBarR(name, ypos, model, parent=self)
            self.rbars.append(bar)

    # Methods added to read from JSON
    @classmethod
    def init_args(self, name, setup_objects):
        barsmodel = setup_objects['barsmodel']
        return (name, barsmodel), {}

    def is_open(self):
        lopen = all(bar.position <= self.OPEN_L_LIMIT for bar in self.lbars)
        ropen = all(bar.position >= self.OPEN_R_LIMIT for bar in self.rbars)

        return lopen and ropen

    def configure_with_header(self, hdr):

        for i in range(self.NBARS):
            csupos = hdr["CSUP{}".format(i + 1)]
            self.lbars[i].csupos = csupos
        for i in range(self.NBARS):
            csupos = hdr["CSUP{}".format(self.NBARS + i + 1)]
            self.lbars[i].csupos = csupos

    @property
    def test(self):
        return 0


class CSUBar(HWDevice):
    def __init__(self, name, ypos, model, parent=None, active=True):
        super(CSUBar, self).__init__(name, parent=parent)
        self.ypos = ypos
        self.csupos_ = 0
        self.active_ = active
        self.model = model
        # Geometry
        self.slit_h_virt = 16.242
        self.y0 = ypos
        self.y1 = ypos - self.slit_h_virt - 3
        self.y2 = ypos + self.slit_h_virt + 3
        self._x1 = 0
        self._x2 = 2047

    @property
    def csupos(self):
        return self.csupos_

    @csupos.setter
    def csupos(self, value):
        if self.active_:
            self.csupos_ = value
        else:
            # warning, bar is inactive
            pass

    @property
    def active(self):
        return self.active_

    @active.setter
    def active(self, value):
        self.active_ = value

    @property
    def position(self):
        return self.model(self.csupos_)

    @property
    def x1(self):
        return self._x1

    @property
    def x2(self):
        return self._x2

    def bbox(self):
        # origin is 1
        x1 = max(self.x1 - 1, 0)
        y1 = max(self.y1 - 1, 0)
        x2 = max(self.x2 - 1, 0, x1)
        y2 = max(self.y2 - 1, 0, y1)
        return BoundingBox.from_coordinates(x1, x2, y1, y2)


class CSUBarL(CSUBar):
    @property
    def x2(self):
        return self.position


class CSUBarR(CSUBar):
    @property
    def x1(self):
        return self.position


def create_test_header0():
    hdr = {}
    for i in range(55):
        hdr["CSUP{}".format(i + 1)] = 123
    for i in range(55, 110):
        hdr["CSUP{}".format(i + 1)] = 125
    return hdr


if __name__ == '__main__':

    import pkgutil
    import numpy
    import io
    import numpy.polynomial.polynomial as pol

    dumdata = pkgutil.get_data('emirdrp.instrument.configs', 'bars_nominal_positions_test.txt')
    ss = io.StringIO(dumdata.decode('utf8'))
    bars_nominal_positions = numpy.loadtxt(ss)

    class BarModel:
        def __init__(self):
            self.values = {}

    barmodel = BarModel()
    barmodel.values['nominal'] = bars_nominal_positions

    csu = ConfigurableSlitUnit("CSU", barmodel)

    print(csu.config_info())
    csu.configure({'CSU.BarL_1': {'csupos': 100}})
    print(csu.children)
    print(csu.is_open())
    dev = csu.get_device('CSU.BarL_12')
    csu.get_property('CSU.BarL_12.csupos')
    dev.csupos = 234
    print(dev.position)
    print(dev.bbox().slice)

    hdr = create_test_header0()

    csu.configure_with_header(hdr)
    print(csu.config_info())



#
# Copyright 2013-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Combination routines"""

from __future__ import division

import datetime
import functools
import logging
import uuid

import numpy
from astropy.io import fits
import numina.datamodel as datam
from numina.frame.utils import copy_img
from numina.array import combine
from numina.array import combine_shape

from emirdrp.processing.wcs import offsets_from_wcs


_logger = logging.getLogger(__name__)


def basic_processing(rinput, flow):
    """

    Parameters
    ----------
    rinput : RecipeInput
    flow

    Returns
    -------

    """
    return basic_processing_(rinput.obresult.images, flow)


def process_abba(images, errors=False, prolog=None):
    """
    Process images in ABBA sequence

    Parameters
    ----------
    images
    errors
    prolog

    Returns
    -------

    """
    cnum = len(images)
    result = copy_img(images[0])
    base_header = result[0].header.copy()

    now = datetime.datetime.utcnow()
    _logger.info("processing ABBA")
    ###
    ###
    dataA0 = images[0][0].data.astype('float32')
    dataB0 = images[1][0].data.astype('float32')
    dataB1 = images[2][0].data.astype('float32')
    dataA1 = images[3][0].data.astype('float32')
    dataAB0 = dataA0 - dataB0
    dataAB1 = dataA1 - dataB1
    data = dataAB0 + dataAB1
    ###
    hdu = result[0]
    hdu.data = data
    _logger.debug('update result header')
    if prolog:
        hdu.header['history'] = prolog
    hdu.header['history'] = "Processed ABBA"
    hdu.header['history'] = 'Combination time {}'.format(now.isoformat())
    for img in images:
        hdu.header['history'] = "Image {}".format(datam.get_imgid(img))
    prevnum = base_header.get('NUM-NCOM', 1)
    hdu.header['NUM-NCOM'] = prevnum * cnum // 2
    hdu.header['UUID'] = str(uuid.uuid1())
    # Headers of last image
    hdu.header['TSUTC2'] = images[-1][0].header['TSUTC2']
    # Not sure why this is needed
    hdu.header['EXTEND'] = True
    for img, key in zip(images, ['A', 'B', 'B', 'A']):
        imgid = datam.get_imgid(img)
        hdu.header['history'] = "Image '{}' is '{}'".format(imgid, key)

    # Update primary extension
    result[0] = hdu

    return result


def process_ab(images, errors=False, prolog=None):
    """
    Process images in AB sequence

    Parameters
    ----------
    images
    errors
    prolog

    Returns
    -------

    """
    cnum = len(images)
    result = copy_img(images[0])
    base_header = result[0].header.copy()
    now = datetime.datetime.utcnow()
    _logger.info("processing AB")
    ###
    ###
    dataA0 = images[0][0].data.astype('float32')
    dataB0 = images[1][0].data.astype('float32')
    data = dataA0 - dataB0

    ###
    hdu = fits.PrimaryHDU(data, header=base_header)
    _logger.debug('update result header')
    if prolog:
        hdu.header['history'] = prolog
    hdu.header['history'] = "Processed AB"
    hdu.header['history'] = 'Combination time {}'.format(now.isoformat())
    for img in images:
        hdu.header['history'] = "Image {}".format(datam.get_imgid(img))
    prevnum = base_header.get('NUM-NCOM', 1)
    hdu.header['NUM-NCOM'] = prevnum * cnum // 2
    hdu.header['UUID'] = str(uuid.uuid1())
    # Headers of last image
    hdu.header['TSUTC2'] = images[-1][0].header['TSUTC2']

    for img, key in zip(images, ['A', 'B']):
        imgid = datam.get_imgid(img)
        hdu.header['history'] = "Image '{}' is '{}'".format(imgid, key)

    # Update primary extension
    # Not sure why this is needed
    hdu.header['EXTEND'] = True
    result[0] = hdu

    return result


def create_proc_hdulist(images, data_array):
    cnum = len(images)
    result = copy_img(images[0])
    base_header = result[0].header.copy()

    hdu = fits.PrimaryHDU(data_array, header=base_header)
    # self.set_base_headers(hdu.header)
    hdu.header['UUID'] = str(uuid.uuid1())
    # Update obsmode in header
    #hdu.header['OBSMODE'] = 'LS_ABBA'
    # Headers of last image
    hdu.header['TSUTC2'] = images[-1][0].header['TSUTC2']
    # Not sure why this is needed
    hdu.header['EXTEND'] = True
    result[0] = hdu
    return result


def basic_processing_(frames, flow):
    """

    Parameters
    ----------
    frames
    flow

    Returns
    -------

    """

    cdata = []
    _logger.info('processing input frames')
    for frame in frames:
        hdulist = frame.open()
        fname = datam.get_imgid(hdulist)
        _logger.info('input is %s', fname)
        final = flow(hdulist)
        _logger.debug('output is input: %s', final is hdulist)

        cdata.append(final)

    return cdata


def scale_with_median(method):
    """Adapt combine method to scale with median
    """
    @functools.wraps(method)
    def wrapper(data_list, dtype='float32'):
        medians = numpy.array([numpy.median(d) for d in data_list])
        scales = medians / medians.max()
        data_com = method(data_list, dtype=dtype, scales=scales)
        return data_com
    # Adapt __name__
    rename = "scaled_" + wrapper.__name__
    wrapper.__name__ = rename

    return wrapper


# FIXME: use the function in numina
def combine_images(images, method=combine.mean, method_kwargs=None,
                   errors=False, prolog=None):
    """Combine a sequence of HDUList objects.

    Using the following keywords:
     * NUM-NCOM
     * UUID, TSUTC1, TSUTC2
     * VARIANCE, MAP

    Parameters
    ----------
    images
    method
    method_kwargs : dict (optional)
    errors : bool (optional)
    prolog

    Returns
    -------

    """
    cnum = len(images)
    result = copy_img(images[0])
    base_header = result[0].header.copy()

    now = datetime.datetime.utcnow()
    _logger.info("stacking %d images using '%s'", cnum, method.__name__)
    data = method([d[0].data for d in images], dtype='float32')
    hdu = result[0]
    hdu.data = data[0]
    _logger.debug('update result header')
    if prolog:
        hdu.header['history'] = prolog
    hdu.header['history'] = "Combined %d images using '%s'" % (cnum, method.__name__)
    hdu.header['history'] = 'Combination time {}'.format(now.isoformat())
    for img in images:
        hdu.header['history'] = "Image {}".format(datam.get_imgid(img))
    prevnum = base_header.get('NUM-NCOM', 1)
    hdu.header['NUM-NCOM'] = prevnum * cnum
    hdu.header['UUID'] = str(uuid.uuid1())
    # Headers of last image
    hdu.header['TSUTC2'] = images[-1][0].header['TSUTC2']
    # Not sure why this is needed
    hdu.header['EXTEND'] = True
    result[0] = hdu
    if errors:
        varhdu = fits.ImageHDU(data[1], name='VARIANCE')
        result.append(varhdu)
        num = fits.ImageHDU(data[2].astype('int16'), name='MAP')
        result.append(num)
    return result


def resize_hdul(hdul, newshape, region, extensions=None, window=None,
                    scale=1, fill=0.0, conserve=True):
    """

    Parameters
    ----------
    hdul
    newshape
    region
    extensions
    window
    scale
    fill
    conserve

    Returns
    -------

    """
    from numina.frame import resize_hdu

    if extensions is None:
        extensions = [0]

    nhdul = [None] * len(hdul)
    for ext, hdu in enumerate(hdul):
        if ext in extensions:
            nhdul[ext] = resize_hdu(hdu, newshape,
                                    region, fill=fill,
                                    window=window,
                                    scale=scale,
                                    conserve=conserve)
        else:
            nhdul[ext] = hdu
    return fits.HDUList(nhdul)


def resize(frames, shape, offsetsp, finalshape, window=None):
    """

    Parameters
    ----------
    frames
    shape
    offsetsp
    finalshape
    window

    Returns
    -------

    """
    from numina.array import subarray_match
    _logger.info('Resizing frames and masks')
    rframes = []
    regions = []
    for frame, rel_offset in zip(frames, offsetsp):
        region, _ = subarray_match(finalshape, rel_offset, shape)
        rframe = resize_hdul(frame.open(), finalshape, region)
        rframes.append(rframe)
        regions.append(region)
    return rframes, regions


def resize_hdulists(hdulists, shape, offsetsp, finalshape, window=None):
    """

    Parameters
    ----------
    hdulists
    shape
    offsetsp
    finalshape
    window

    Returns
    -------

    """
    from numina.array import subarray_match
    _logger.info('Resizing frames and masks')
    rhdulist = []
    regions = []
    for hdulist, rel_offset in zip(hdulists, offsetsp):
        region, _ = subarray_match(finalshape, rel_offset, shape)
        rframe = resize_hdul(hdulist, finalshape, region)
        rhdulist.append(rframe)
        regions.append(region)
    return rhdulist, regions


def basic_processing_with_segmentation(rinput, flow,
                                          method=combine.mean,
                                          errors=True, bpm=None):

    odata = []
    cdata = []
    try:
        _logger.info('processing input images')
        for frame in rinput.obresult.images:
            hdulist = frame.open()
            fname = datam.get_imgid(hdulist)
            _logger.info('input is %s', fname)
            final = flow(hdulist)
            _logger.debug('output is input: %s', final is hdulist)

            cdata.append(final)

            # Files to be closed at the end
            odata.append(hdulist)
            if final is not hdulist:
                odata.append(final)

        base_header = cdata[0][0].header.copy()

        baseshape = cdata[0][0].data.shape
        subpixshape = cdata[0][0].data.shape

        _logger.info('Computing offsets from WCS information')
        refpix = numpy.divide(numpy.array([baseshape], dtype='int'), 2).astype('float')
        offsets_xy = offsets_from_wcs(rinput.obresult.frames, refpix)
        _logger.debug("offsets_xy %s", offsets_xy)
        # Offsets in numpy order, swaping
        offsets_fc = offsets_xy[:, ::-1]
        offsets_fc_t = numpy.round(offsets_fc).astype('int')

        _logger.info('Computing relative offsets')
        finalshape, offsetsp = combine_shape(subpixshape, offsets_fc_t)
        _logger.debug("offsetsp %s", offsetsp)

        _logger.info('Shape of resized array is %s', finalshape)
        # Resizing target frames
        rhduls, regions = resize_hdulists(cdata, subpixshape, offsetsp, finalshape)

        _logger.info("stacking %d images, with offsets using '%s'", len(cdata), method.__name__)
        data1 = method([d[0].data for d in rhduls], dtype='float32')

        segmap = segmentation_combined(data1[0])
        # submasks
        if bpm is None:
            masks = [(segmap[region] > 0) for region in regions]
        else:
            masks = [((segmap[region] > 0) & bpm) for region in regions]

        _logger.info("stacking %d images, with objects mask using '%s'", len(cdata), method.__name__)
        data2 = method([d[0].data for d in cdata], masks=masks, dtype='float32')
        hdu = fits.PrimaryHDU(data2[0], header=base_header)
        points_no_data = (data2[2] == 0).sum()

        _logger.debug('update result header')
        hdu.header['TSUTC2'] = cdata[-1][0].header['TSUTC2']
        hdu.header['history'] = "Combined %d images using '%s'" % (len(cdata), method.__name__)
        hdu.header['history'] = 'Combination time {}'.format(datetime.datetime.utcnow().isoformat())
        hdu.header['UUID'] = str(uuid.uuid1())
        _logger.info("missing points, total: %d, fraction: %3.1f", points_no_data, points_no_data / data2[2].size)

        if errors:
            varhdu = fits.ImageHDU(data2[1], name='VARIANCE')
            num = fits.ImageHDU(data2[2], name='MAP')
            result = fits.HDUList([hdu, varhdu, num])
        else:
            result = fits.HDUList([hdu])
    finally:
        _logger.debug('closing images')
        for hdulist in odata:
            hdulist.close()

    return result


def segmentation_combined(data, snr_detect=10.0, fwhm=4.0, npixels=15, mask_corners=False):
    import sep
    from astropy.convolution import Gaussian2DKernel
    from astropy.stats import gaussian_fwhm_to_sigma

    box_shape = [64, 64]
    _logger.info('point source detection')

    # Corners
    mask = numpy.zeros_like(data, dtype='int32')
    if mask_corners:
        # Remove corner regions
        _logger.info('using internal mask to remove corners')
        mask[2000:, 0:80] = 1
        mask[2028:, 2000:] = 1
        mask[:50, 1950:] = 1
        mask[:100, :50] = 1

    _logger.info('compute background map, %s', box_shape)
    bkg = sep.Background(data)

    _logger.info('reference fwhm is %5.1f pixels', fwhm)

    _logger.info('convolve with gaussian kernel, FWHM %3.1f pixels', fwhm)
    sigma = fwhm * gaussian_fwhm_to_sigma
    #
    kernel = Gaussian2DKernel(sigma)
    kernel.normalize()

    _logger.info('background level is %5.1f', bkg.globalback)
    _logger.info('background rms is %5.1f', bkg.globalrms)
    _logger.info('detect threshold, %3.1f sigma over background', snr_detect)
    thresh = snr_detect * bkg.globalrms

    data_s = data - bkg.back()
    try:
        objects, segmap = sep.extract(data_s, thresh, minarea=npixels,
                                      filter_kernel=kernel.array, segmentation_map=True,
                                      mask=mask)
        _logger.info('detected %d objects', len(objects))
    except Exception as error:
        _logger.warning("%s", error)
        segmap = numpy.zeros_like(data_s, dtype='int')
    return segmap


def basic_processing_with_update(rinput, flow):

    # FIXME: this only works with local images
    # We don't know how to store temporary GCS frames
    _logger.info('processing input images')
    for frame in rinput.obresult.images:
        with fits.open(frame.label, mode='update') as hdul:
            flow(hdul)




import pytest
import astropy.io.fits as fits
from astropy import wcs
from astropy.modeling.functional_models import Gaussian2D
import numpy
import numpy.random

import numina.core
import numina.exceptions

from ..join import JoinDitheredImagesRecipe


def create_wcs(crpix=None):

    w = wcs.WCS(naxis=2)

    crpix = [50.0, 50.0] if crpix is None else crpix

    w.wcs.crpix = crpix
    w.wcs.crval = [0.0, 0.0]
    w.wcs.cdelt = [1.0, 1.0]
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    w.wcs.cd = numpy.array([
        [5.41615466e-05, -2.55232165e-07],
        [2.55232165e-07,  5.41615466e-05]
    ])
    return w


def dither_pattern(center, base_angle, dist, npoints):

    base_angle_rad = base_angle / 180.0 * numpy.pi
    step = 2*numpy.pi / npoints
    angles = base_angle_rad + numpy.arange(0.0, 2 * numpy.pi, step)
    x = center[0] + dist * numpy.cos(angles)
    y = center[1] + dist * numpy.sin(angles)
    return numpy.asarray([x, y]).T


def create_frame(background=0, std=100, crpix=None, keys=None):

    size = (100, 100)
    # Update WCS header
    wcs = create_wcs(crpix)
    y, x = numpy.mgrid[:100, :100]

    data = numpy.random.normal(background, std, size)
    model = Gaussian2D(
        amplitude=30*background,
        x_mean=wcs.wcs.crpix[0]-3.5,
        y_mean=wcs.wcs.crpix[1]-12.8,
        x_stddev=3.0,
        y_stddev=4.0
    )
    data += model(x, y)
    hdu = fits.PrimaryHDU(data)

    hdu.header = wcs.to_header()
    if keys:
        for k, v in keys.items():
            hdu.header[k] = v

    return fits.HDUList([hdu])


def create_ob(value, nimages, crpix, nstare=3, exptime=100.0, starttime=0.0):

    frames = []
    for i in range(nimages):
        base = starttime
        off1 = i * exptime
        off2 = off1 + exptime
        t1 = base + off1
        t2 = base + off2
        keys = {'TSUTC1': t1, 'TSUTC2': t2, 'NUM-NCOM': nstare, 'NUM-SK': 1}
        frame = numina.core.DataFrame(frame=create_frame(background=value, keys=keys, crpix=crpix[i]))
        frames.append(frame)

    obsresult = numina.core.ObservationResult()
    obsresult.frames = frames
    return obsresult


@pytest.mark.parametrize("nimages,naccum", [(7, 10)])
def test_join(nimages, naccum):
    nstare = 3
    value = 13.0
    inittime = 1030040001.00034
    starttime = inittime
    exptime = 105.0
    crpix = dither_pattern([50, 50], 0.0, 20.0, nimages)
    obsresult = create_ob(value, nimages, crpix, nstare, exptime, starttime)
    obsresult.naccum = 1

    recipe = JoinDitheredImagesRecipe()
    prev_accum = None

    while True:

        rinput = recipe.create_input(
            obresult=obsresult,
            accum=prev_accum
        )

        result = recipe(rinput)
        frame_hdul = result.frame.open()
        assert frame_hdul[0].header['NUM-NCOM'] == nimages * nstare
        accum_hdul = result.accum.open()
        assert accum_hdul[0].header['NUM-NCOM'] == nimages * nstare * obsresult.naccum
        if obsresult.naccum < naccum:
            # Init next loop
            starttime += exptime * (nimages + 1)
            nobsresult = create_ob(value , nimages, crpix, nstare, exptime, starttime)
            nobsresult.naccum = obsresult.naccum + 1
            obsresult = nobsresult
            prev_accum = result.accum
        else:
            break

    frame_hdul = result.frame.open()
    accum_hdul = result.accum.open()

    assert frame_hdul[0].header['NUM-NCOM'] == nimages * nstare
    assert frame_hdul['MAP'].data.max() == nimages
    assert frame_hdul[0].header['TSUTC1'] == starttime
    assert frame_hdul[0].header['TSUTC2'] == starttime + nimages * exptime
    assert accum_hdul[0].header['NUM-NCOM'] == nimages * nstare * naccum
    assert accum_hdul[0].header['TSUTC1'] == inittime
    assert accum_hdul[0].header['TSUTC2'] == starttime + nimages * exptime


@pytest.mark.parametrize("nimages", [2, 3, 7])
def test_init_aggregate_result(nimages):
    # Test that works in the first run of the loop
    nstare = 3
    value = 9
    starttime = 1030040001.00034
    exptime = 105.0
    crpix = numpy.array([(50, 50) for _ in range(nimages)])
    obsresult = create_ob(value, nimages, crpix, nstare, exptime, starttime)
    obsresult.naccum = 1
    recipe = JoinDitheredImagesRecipe()
    rinput = recipe.create_input(
        obresult=obsresult,
    )

    #import logging
    #logging.basicConfig(level=logging.DEBUG)
    result1 = recipe.run(rinput)

    frame_hdul = result1.frame.open()
    accum_hdul = result1.accum.open()

    assert(len(frame_hdul) == len(accum_hdul))

    for key in ['NUM-NCOM', 'TSUTC1', 'TSUTC2']:
        assert frame_hdul[0].header[key] == accum_hdul[0].header[key]

    assert frame_hdul[0].header['NUM-NCOM'] == nimages * nstare

    assert numpy.allclose(frame_hdul[0].data, accum_hdul[0].data)



#
# Copyright 2008-2014 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#


def gather_info_dframe(dataframe):
    with dataframe.open() as hdulist:
        info = gather_info_hdu(hdulist)
    return info


def gather_info_hdu(hdulist):

    _meta = {'readmode': ('READMODE', 'undefined'),
             'texp': ('EXPTIME', None),
             'grism': ('GRISM', 'undefined'),
             'filter': ('FILTER', 'undefined'),
             'obsmode': ('OBSMODE', 'undefined'),
             'tstamp': ('TSTAMP', 'undefined')
             }

    # READMODE is STRING
    meta = {}
    meta['n_ext'] = len(hdulist)
    extnames = [hdu.header.get('extname', '') for hdu in hdulist[1:]]
    meta['name_ext'] = ['PRIMARY'] + extnames
    for key, val in _meta.items():
        meta[key] = hdulist[0].header.get(val[0], val[1])

    return meta


def gather_info_frames(framelist):
    iinfo = []
    for frame in framelist:
        with frame.open() as hdulist:
            iinfo.append(gather_info_hdu(hdulist))
    return iinfo


def gather_info(recipeinput):
    from numina.core import DataFrame
    klass = recipeinput.__class__
    metadata = {}
    for key in klass.stored():
        val = getattr(recipeinput, key)
        if isinstance(val, DataFrame):
            metadata[key] = gather_info_dframe(val)
        elif hasattr(val, 'frames'):
            metas = []
            for f in val.frames:
                metas.append(gather_info_dframe(f))
            metadata[key] = metas
        else:
            pass
    return metadata



#
# Copyright 2012-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Micro-dithering Recipe of EMIR

"""

from numina.core import Parameter
from numina.core.recipeinout import define_input, define_result
from numina.core import Result, DataFrameType, RecipeInput, RecipeResult
from numina.core.requirements import ObservationResultRequirement

from emirdrp.products import SourcesCatalog
from emirdrp.requirements import Offsets_Requirement
from emirdrp.requirements import SkyImageSepTime_Requirement
from emirdrp.requirements import MasterBadPixelMaskRequirement
from emirdrp.requirements import MasterBiasRequirement
from emirdrp.requirements import MasterDarkRequirement
from emirdrp.requirements import MasterIntensityFlatFieldRequirement

from .shared import DirectImageCommon


class MicroditheredImageRecipeInput(RecipeInput):
    obresult = ObservationResultRequirement()
    master_bpm = MasterBadPixelMaskRequirement()
    master_bias = MasterBiasRequirement()
    master_dark = MasterDarkRequirement()
    master_flat = MasterIntensityFlatFieldRequirement()
    extinction = Parameter(0.0, 'Mean atmospheric extinction')
    sources = Parameter([],
                        'List of x, y coordinates to measure FWHM',
                        optional=True)
    offsets = Offsets_Requirement()
    iterations = Parameter(4, 'Iterations of the recipe')
    sky_images = Parameter(5, 'Images used to estimate the background before '
                           'and after current image')
    sky_images_sep_time = SkyImageSepTime_Requirement()
    check_photometry_levels = Parameter(
        [0.5, 0.8], 'Levels to check the flux of the objects')
    check_photometry_actions = Parameter(
        ['warn', 'warn', 'default'], 'Actions to take on images')
    subpixelization = Parameter(4, 'Number of subdivisions in each pixel side')
    window = Parameter([], 'Region of interesting data', optional=True)


class MicroditheredImageRecipeResult(RecipeResult):
    frame = Result(DataFrameType)
    catalog = Result(SourcesCatalog)


@define_input(MicroditheredImageRecipeInput)
@define_result(MicroditheredImageRecipeResult)
class MicroditheredImageRecipe(DirectImageCommon):

    """
    Recipe for the reduction of microdithering imaging.

    The effect of recording a series of stare images, with the same
    acquisition parameters, and taken by pointing to a number of
    sky positions, with separations of the order of sub arcsecs,
    either by moving the either by nodding the TS, tilting the TS
    M2 or shifting the EMIR DTU, the latter being the most likely
    option. Displacements are of the order of fraction of pixels.
    Images share the large majority of the sky positions so they can
    be coadded. Used for improving the spatial resolution of the
    resulting images and not valid for sky or superflat images.


    **Observing modes:**

        * Micro-dithered images


    Recipe to reduce observations obtained in imaging mode with microdithering.
    A critical piece of information
    here is a table that clearly specifies which images can be labelled as
    *science*, and which ones as *sky*. Note that some images are used both as
    *science* and *sky* (when the size of the targets are small compared to the
    offsets).

    **Observing modes:**
     * Micro-dithered images

    **Inputs:**


        * Offsets between them
        * Master Dark
        * Bad pixel mask (BPM)
        * Non-linearity correction polynomials
        * Master flat (twilight/dome flats)
        * Master background (thermal background, only in K band)
        * Detector model (gain, RN, lecture mode)
        * Average extinction in the filter
        * Astrometric calibration (TBD)

    **Outputs:**

     * Image with three extensions: final image scaled to the individual
       exposure time, variance  and exposure time map OR number of images
       combined (TBD).

    **Procedure:**

    Images are regridded to a integer subdivision of the pixel and then they
    are corrected from dark, non-linearity and flat. It should be desirable
    that the microdithering follows a pattern that can be easily translated to
    a subdivision of the pixel size (by an integer *n* = 2, 3, 4,...) that
    does not require a too high *n* value. An iterative process starts:

     * Sky is computed from each frame, using the list of sky images of each
       science frame. The objects are avoided using a mask (from the second
       iteration on).

     * The relatiev offsets are the nominal from the telescope. From the second
       iteration on, we refine them using bright objects.

     * We combine the sky-subtracted images, output is: a new image, a variance
       image and a exposure map/number of images used map.

     * An object mask is generated.

     * We recompute the sky map, using the object mask as an additional input.
       From here we iterate (typically 4 times).

     * Finally, the images are corrected from atmospheric extinction and flux
       calibrated.

     * A preliminary astrometric calibration can always be used (using
       the central coordinates of the pointing and the plate
       scale in the detector).
       A better calibration might be computed using available stars (TBD).

    """

    def run(self, ri):

        subpix = ri.subpixelization
        window = ri.window

        frame, catalog = self.process(ri,
                                      window=window, subpix=subpix,
                                      target_is_sky=True,
                                      stop_after=DirectImageCommon.FULLRED)

        result = self.create_result(frame=frame, catalog=catalog)
        return result



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
import json
import numpy as np
from numpy.polynomial import Polynomial
import os.path
import sys

from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximshow import ximshow

from emirdrp.core import EMIR_NBARS
from emirdrp.core import EMIR_NAXIS1


def get_boundaries(bounddict_file, slitlet_number):
    """Read the bounddict json file and return the polynomial boundaries.

    Parameters
    ----------
    bounddict_file : file handler
        File containing the bounddict JSON data.
    slitlet_number : int
        Number of slitlet.

    Returns
    -------
    pol_lower_boundary : numpy polynomial
        Polynomial defining the lower boundary of the slitlet.
    pol_upper_boundary : numpy polynomial
        Polynomial defining the upper boundary of the slitlet.
    xmin_lower : float
        Minimum abscissae for the lower boundary.
    xmax_lower : float
        Maximum abscissae for the lower boundary.
    xmin_upper : float
        Minimum abscissae for the upper boundary.
    xmax_upper : float
        Maximum abscissae for the upper boundary.
    csu_bar_slit_center : float
        CSU bar slit center (in mm)

    """

    bounddict = json.loads(open(bounddict_file.name).read())

    # return values in case the requested slitlet number is not defined
    pol_lower_boundary = None
    pol_upper_boundary = None
    xmin_lower = None
    xmax_lower = None
    xmin_upper = None
    xmax_upper = None
    csu_bar_slit_center = None

    # search the slitlet number in bounddict
    slitlet_label = "slitlet" + str(slitlet_number).zfill(2)
    if slitlet_label in bounddict['contents'].keys():
        list_date_obs = list(bounddict['contents'][slitlet_label].keys())
        list_date_obs.sort()
        num_date_obs = len(list_date_obs)
        if num_date_obs == 1:
            date_obs = list_date_obs[0]
            tmp_dict = bounddict['contents'][slitlet_label][date_obs]
            pol_lower_boundary = Polynomial(tmp_dict['boundary_coef_lower'])
            pol_upper_boundary = Polynomial(tmp_dict['boundary_coef_upper'])
            xmin_lower = tmp_dict['boundary_xmin_lower']
            xmax_lower = tmp_dict['boundary_xmax_lower']
            xmin_upper = tmp_dict['boundary_xmin_upper']
            xmax_upper = tmp_dict['boundary_xmax_upper']
            csu_bar_slit_center = tmp_dict['csu_bar_slit_center']
        else:
            raise ValueError("num_date_obs =", num_date_obs,
                             " (must be 1)")
    else:
        print("WARNING: slitlet number " + str(slitlet_number) +
              " is not available in " + bounddict_file.name)

    # return result
    return pol_lower_boundary, pol_upper_boundary, \
           xmin_lower, xmax_lower, xmin_upper, xmax_upper, \
           csu_bar_slit_center


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()

    # positional arguments
    parser.add_argument("fitsfile",
                        help="FITS file name to be displayed",
                        type=argparse.FileType('rb'))
    parser.add_argument("--bounddict", required=True,
                        help="bounddict file name",
                        type=argparse.FileType('rt'))
    parser.add_argument("--tuple_slit_numbers", required=True,
                        help="Tuple n1[,n2[,step]] to define slitlet numbers")

    # optional arguments
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args()

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # read slitlet numbers to be computed
    tmp_str = args.tuple_slit_numbers.split(",")
    if len(tmp_str) == 3:
        if int(tmp_str[0]) < 1:
            raise ValueError("Invalid slitlet number < 1")
        if int(tmp_str[1]) > EMIR_NBARS:
            raise ValueError("Invalid slitlet number > EMIR_NBARS")
        list_slitlets = range(int(tmp_str[0]),
                              int(tmp_str[1])+1,
                              int(tmp_str[2]))
    elif len(tmp_str) == 2:
        if int(tmp_str[0]) < 1:
            raise ValueError("Invalid slitlet number < 1")
        if int(tmp_str[1]) > EMIR_NBARS:
            raise ValueError("Invalid slitlet number > EMIR_NBARS")
        list_slitlets = range(int(tmp_str[0]),
                              int(tmp_str[1])+1,
                              1)
    elif len(tmp_str) == 1:
        if int(tmp_str[0]) < 1:
            raise ValueError("Invalid slitlet number < 1")
        if int(tmp_str[0]) > EMIR_NBARS:
            raise ValueError("Invalid slitlet number > EMIR_NBARS")
        list_slitlets = [int(tmp_str[0])]
    else:
        raise ValueError("Invalid tuple for slitlet numbers")

    # read input FITS file
    hdulist = fits.open(args.fitsfile.name)
    image_header = hdulist[0].header
    image2d = hdulist[0].data
    hdulist.close()

    naxis1 = image_header['naxis1']
    naxis2 = image_header['naxis2']

    if image2d.shape != (naxis2, naxis1):
        raise ValueError("Unexpected error with NAXIS1, NAXIS2")

    # remove path from fitsfile
    sfitsfile = os.path.basename(args.fitsfile.name)

    # check that the FITS file has been obtained with EMIR
    instrument = image_header['instrume']
    if instrument != 'EMIR':
        raise ValueError("INSTRUME keyword is not 'EMIR'!")

    # read GRISM, FILTER and ROTANG from FITS header
    grism = image_header['grism']
    spfilter = image_header['filter']
    rotang = image_header['rotang']

    # display full image
    ax = ximshow(image2d=image2d,
                 title=sfitsfile + "\ngrism=" + grism +
                       ", filter=" + spfilter +
                       ", rotang=" + str(round(rotang, 2)),
                 image_bbox=(1, naxis1, 1, naxis2), show=False)

    # overplot boundaries for each slitlet
    for slitlet_number in list_slitlets:
        pol_lower_boundary, pol_upper_boundary, \
        xmin_lower, xmax_lower, xmin_upper, xmax_upper,  \
        csu_bar_slit_center = \
            get_boundaries(args.bounddict, slitlet_number)
        if (pol_lower_boundary is not None) and \
                (pol_upper_boundary is not None):
            xp = np.linspace(start=xmin_lower, stop=xmax_lower, num=1000)
            yp = pol_lower_boundary(xp)
            ax.plot(xp, yp, 'g-')
            xp = np.linspace(start=xmin_upper, stop=xmax_upper, num=1000)
            yp = pol_upper_boundary(xp)
            ax.plot(xp, yp, 'b-')
            # slitlet label
            yc_lower = pol_lower_boundary(EMIR_NAXIS1 / 2 + 0.5)
            yc_upper = pol_upper_boundary(EMIR_NAXIS1 / 2 + 0.5)
            tmpcolor = ['r', 'b'][slitlet_number % 2]
            xcsu = EMIR_NAXIS1 * csu_bar_slit_center/341.5
            ax.text(xcsu, (yc_lower + yc_upper) / 2,
                    str(slitlet_number),
                    fontsize=10, va='center', ha='center',
                    bbox=dict(boxstyle="round,pad=0.1",
                              fc="white", ec="grey"),
                    color=tmpcolor, fontweight='bold',
                    backgroundcolor='white')

    # show plot
    pause_debugplot(12, pltshow=True)


if __name__ == "__main__":

    main()



#
# Copyright 2016-2017 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Offsets from cross-correlation"""


from __future__  import print_function

import logging

import numpy
import numpy.linalg
import scipy.signal
import numina.array.imsurfit as imsurfit
import numina.array.utils as utils
import numina.array.stats as s


_logger = logging.getLogger('numina.recipes.emir')


def standarize(arr):
    m = arr.mean()
    s = arr.std()
    return (arr - m) / s


def vertex_of_quadratic(coeffs):
    C = coeffs[1]
    D = coeffs[2]
    A = coeffs[3]
    E = coeffs[4]
    B = coeffs[5]

    det = 4 * A * B - E**2
    if det <= 0:
        raise ValueError('quadratic has no maximum')

    xm = -(2*B*C- D*E) / det
    ym = -(2*A*D- C*E) / det
    return xm, ym


def filter_region(arr, level=4):
    median = numpy.median(arr)
    std = s.robust_std(arr, debug=False)

    return numpy.where(arr >= median + level * std, arr, 0.0)


def offsets_from_crosscor(arrs, region, refine=True, refine_box=3, order='ij'):
    # import astropy.io.fits as fits
    # allowed values for order
    if order not in ['xy', 'ij']:
        raise ValueError("'order' must be either 'ij' or 'xy'")

    result = numpy.zeros((len(arrs), 2))
    ref_array = arrs[0]

    for idx, arr in enumerate(arrs[1:], 1):

        refoff = offset_from_crosscor(ref_array, arr, region,
                                      refine=refine,
                                      refine_box=refine_box,
                                      order=order
                                      )
        result[idx] = refoff

    return result


def offsets_from_crosscor_regions(arrs, regions, refine=True, refine_box=3, order='ij', tol=0.5):
    # import astropy.io.fits as fits
    # allowed values for order
    if order not in ['xy', 'ij']:
        raise ValueError("'order' must be either 'ij' or 'xy'")

    result = numpy.zeros((len(arrs), 2))
    ref_array = arrs[0]

    for idx, arr in enumerate(arrs[1:], 1):

        refoff = offset_from_crosscor_regions(
            ref_array, arr, regions,
            refine=refine,
            refine_box=refine_box,
            order=order,
            tol=tol
        )
        result[idx] = refoff

    return result


def offset_from_crosscor(arr0, arr1, region, refine=True, refine_box=3, order='ij'):
    # import astropy.io.fits as fits
    # allowed values for order
    if order not in ['xy', 'ij']:
        raise ValueError("'order' must be either 'ij' or 'xy'")

    ref_array = arr0
    shape = ref_array.shape
    data1 = filter_region(ref_array[region])
    d1 = standarize(data1)
    # fits.writeto('cutout_%d.fits' % 0, d1, overwrite=True)

    dcenter = numpy.asarray(d1.shape) // 2
    # Correlate

    data2 = filter_region(arr1[region])
    d2 = standarize(data2)
    # fits.writeto('cutout_%d.fits' % idx, d2, overwrite=True)
    #corr = scipy.signal.correlate2d(d1, d2, mode='same', boundary='fill', fillvalue=fillvalue)
    # correlation is equivalent to convolution with inverted image
    corr = scipy.signal.fftconvolve(d1, d2[::-1, ::-1], mode='same')
    # normalize
    corr /= corr.max()
    # fits.writeto('corr_%d.fits' % idx, corr, overwrite=True)
    # fits.writeto('corr2_%d.fits' % idx, corr2 / corr2.max(), overwrite=True)
    # Find peak in cross-cor
    maxindex = numpy.unravel_index(corr.argmax(), corr.shape)

    # Check the peak is above n times the background
    median_corr = numpy.median(corr)
    std_corr = numpy.std(corr)
    peakvalue = corr[maxindex]
    threshold = median_corr + 5 * std_corr
    _logger.debug('The peak value is %f, threshold %f', peakvalue, threshold)
    if peakvalue < threshold:
        raise ValueError('Peak below threshold')

    #baseoff = dcenter - numpy.asarray(maxindex)
    # Pixel (0,0) in reference corresponds to baseoff in image
    # print("Pixel (0,0) in reference corresponds to %s in image" % baseoff)

    if refine:
        # Refine to subpixel
        # Fit a 2D surface to the peak of the crosscorr
        region_ref = utils.image_box(maxindex, shape, box=(refine_box, refine_box))

        coeffs, = imsurfit.imsurfit(corr[region_ref], order=2)
        # coefss are a + b *x + c * y + d*x**2 + e * x* y + f * y**2

        try:
            # get peak from coeffs
            xm, ym = vertex_of_quadratic(coeffs)
            # xm ym are offsets

            if abs(xm) > 1 or abs(ym) > 1:
                # probably bad fit
                # dont apply
                final = maxindex
            else:
                final = maxindex + numpy.asarray([ym, xm])
        except ValueError as error:
            _logger.debug('Error fitting peak, %s', error)
            final = maxindex
    else:
        final = maxindex

    refoff = dcenter - final
    # Pixel (0,0) in reference corresponds to baseoff in image
    # print("Pixel (0,0) in reference corresponds to %s in image" % refoff)
    # result xy
    if order == 'xy':
        return refoff[::-1]
    else:
        return refoff


def offset_from_crosscor_regions(arr0, arr1, regions, refine=True, refine_box=3, order='ij', tol=0.5):

    values = []
    for region in regions:
        try:
            res = offset_from_crosscor(arr0, arr1, region,
                                       refine=refine,
                                       refine_box=refine_box,
                                       order=order)
            values.append(res)
        except ValueError as error:
            _logger.debug('error in offset_from_crosscor_regions: %s', error)

    if len(values) == 0:
        raise ValueError('No measurements to compute offset in any region')
    values = numpy.array(values)
    _logger.debug('offsets from cross-correlation:\n%s', values)
    med_c = numpy.median(values, axis=0)

    dis = numpy.linalg.norm(values - med_c, axis=1)
    # filter values with dis > tol
    mask = dis <= tol
    # Compute mean
    if numpy.any(mask):
        return values[mask].mean(axis=0)
    else:
        raise ValueError('No measurements within tolerance in any region')



#
# Copyright 2016-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Spectroscopy mode, ABBA
"""
import uuid
import pkgutil

import astropy.io.fits as fits
import numpy

import numina.core
import numina.exceptions
import numina.core.query as qmod
import numina.ext.gtc
from numina.array import combine
from numina.core import Result, Requirement
from numina.exceptions import RecipeError
from numina.core.requirements import ObservationResultRequirement
from numina.frame.utils import copy_img

import emirdrp.datamodel
import emirdrp.decorators
import emirdrp.products as prods
from emirdrp.core.recipe import EmirRecipe
from emirdrp.processing.combine import basic_processing


class BaseABBARecipe(EmirRecipe):
    """Process images in ABBA mode"""

    obresult = ObservationResultRequirement(
        query_opts=qmod.ResultOf(
            'STARE_SPECTRA.reduced_mos',
            node='children',
            id_field="stareSpectraIds"
        )
    )
    accum_in = Requirement(
        prods.DataFrameType,
        description='Accumulated result',
        optional=True,
        destination='accum',
        query_opts=qmod.ResultOf(
            'LS_ABBA.accum',
            node='prev'
        )
    )

    reduced_mos_abba = Result(prods.ProcessedMOS)
    # Accumulate 'reduced_mos_abba' results
    accum = Result(prods.ProcessedMOS, optional=True)

    def build_recipe_input(self, obsres, dal, pipeline='default'):
        if numina.ext.gtc.check_gtc():
            self.logger.debug('running in GTC environment')
            return self.build_recipe_input_gtc(obsres, dal)
        else:
            self.logger.debug('running outside of GTC environment')
            return super(BaseABBARecipe, self).build_recipe_input(
                obsres, dal
            )

    def build_recipe_input_gtc(self, obsres, dal):
        self.logger.debug('start recipe input builder')
        stareImagesIds = obsres.stareSpectraIds
        self.logger.debug('Stare Spectra images IDS: %s', stareImagesIds)
        stareImages = []
        for subresId in stareImagesIds:
            subres = dal.getRecipeResult(subresId)
            stareImages.append(subres['elements']['reduced_mos'])

        naccum = obsres.naccum
        self.logger.info('naccum: %d', naccum)
        if naccum != 1:  # if it is not the first dithering loop
            self.logger.info("SEARCHING LATEST RESULT LS_ABBA TO ACCUMULATE")
            latest_result = dal.getLastRecipeResult("EMIR", "EMIR", "LS_ABBA")
            accum_dither = latest_result['elements']['accum']
            self.logger.info("FOUND")
        else:
            self.logger.info("NO ACCUMULATION LS_ABBA")
            accum_dither = stareImages[0]

        newOR = numina.core.ObservationResult()
        newOR.frames = stareImages
        newOR.naccum = naccum
        newOR.accum = accum_dither
        newRI = self.create_input(obresult=newOR)
        self.logger.debug('end recipe input builder')
        return newRI

    def run(self, rinput):
        partial_result = self.run_single(rinput)
        new_result = self.aggregate_result(partial_result, rinput)
        return new_result

    def run_single(self, rinput):
        self.logger.info('starting spectroscopy ABBA reduction')

        flow = self.init_filters(rinput)
        nimages = len(rinput.obresult.frames)
        self.logger.info('we receive %d images', nimages)
        if nimages != 4:
            msg = 'Recipe expects 4 images, received %d' % nimages
            raise numina.exceptions.RecipeError(msg)

        procesed_hdulists = basic_processing(rinput, flow)

        # INPUTS are ABBA, so
        #
        hdulist = self.process_abba(procesed_hdulists)
        grism = hdulist[0].header.get('GRISM', 'unknown')
        if grism.lower() == 'open':
            # perform TEST10 in addition
            import emirdrp.recipes.acquisition.maskcheck as mk
            from numina.core import ObservationResult, DataFrame
            import numpy
            try:
                import StringIO as S
            except ImportError:
                import io as S

            self.logger.info('GRISM is OPEN, doing a RECIPE10')
            sub = mk.MaskCheckRecipe()
            sub.configure(instrument='EMIR', mode='TEST10')
            o = ObservationResult()
            o.__dict__ = rinput.obresult.__dict__
            o.frames = [DataFrame(frame=hdulist)]
            subd = {}
            subd['obresult'] = o
            bardata = pkgutil.get_data('emirdrp.instrument.configs', 'bars_nominal_positions_test.txt')
            ss = S.StringIO(bardata.decode('utf8'))
            subd['bars_nominal_positions'] = numpy.loadtxt(ss)

            subinput = mk.MaskCheckRecipe.RecipeInput(**subd)

            sub.run(subinput)

        result = self.create_result(reduced_mos_abba=hdulist)
        self.logger.info('end spectroscopy ABBA reduction')
        return result

    def process_abba(self, images):
        """Process four images in ABBA mode"""
        # FIXME: Duplicated in processing
        dataA0 = images[0][0].data.astype('float32')
        dataB0 = images[1][0].data.astype('float32')

        dataB1 = images[2][0].data.astype('float32')
        dataA1 = images[3][0].data.astype('float32')

        dataAB0 = dataA0 - dataB0
        dataAB1 = dataA1 - dataB1

        dataABBA = dataAB0 + dataAB1

        hdulist = self.create_proc_hdulist(images, dataABBA)
        self.logger.debug('update result header')
        hdu = hdulist[0]
        hdu.header['history'] = "Processed ABBA"
        hdu.header['NUM-NCOM'] = (2, 'Number of combined frames')
        hdu.header['IMGOBBL'] = 0
        dm = emirdrp.datamodel.EmirDataModel()
        for img, key in zip(images, ['A', 'B', 'B', 'A']):
            imgid = dm.get_imgid(img)
            hdu.header['history'] = "Image '{}' is '{}'".format(imgid, key)

        return hdulist

    def create_proc_hdulist(self, cdata, data_array):
        # Copy header of first image
        result = copy_img(cdata[0])

        hdu = result[0]
        hdu.data = data_array
        self.set_base_headers(hdu.header)
        hdu.header['UUID'] = str(uuid.uuid1())
        # Update obsmode in header
        hdu.header['OBSMODE'] = 'LS_ABBA'
        # Headers of last image
        hdu.header['TSUTC2'] = cdata[-1][0].header['TSUTC2']
        return result

    def create_accum_hdulist(self, cdata, data_array_n,
                             method_name='unkwnow', use_errors=False):
        # FIXME: duplicated
        result = copy_img(cdata[0])
        hdu = result[0]
        hdu.data = data_array_n[0]
        hdr = hdu.header
        self.set_base_headers(hdr)
        hdu.header['UUID'] = str(uuid.uuid1())
        hdr['IMGOBBL'] = 0
        hdr['TSUTC2'] = cdata[-1][0].header['TSUTC2']

        hdu.header['history'] = "Combined %d images using '%s'" % (
            len(cdata),
            method_name
        )
        #hdu.header['history'] = 'Combination time {}'.format(
        #    datetime.datetime.utcnow().isoformat()
        #)
        # Update NUM-NCOM, sum of individual frames
        ncom = 0
        for hdul in cdata:
            ncom += hdul[0].header['NUM-NCOM']
        hdr['NUM-NCOM'] = ncom
        #
        if use_errors:
            varhdu = fits.ImageHDU(data_array_n[1], name='VARIANCE')
            result.append(varhdu)
            num = fits.ImageHDU(data_array_n[2], name='MAP')
            result.append(num)
        return result

    def aggregate_result(self, partial_result, rinput):
        obresult = rinput.obresult
        # Check if this is our first run
        naccum = getattr(obresult, 'naccum', 0)
        accum = getattr(obresult, 'accum', None)
        # result to accumulate
        result_key = 'reduced_mos_abba'
        field_to_accum = getattr(partial_result, result_key)

        if naccum == 0:
            self.logger.debug('naccum is not set, do not accumulate')
            return partial_result
        elif naccum == 1:
            self.logger.debug('round %d initialize accumulator', naccum)
            newaccum = field_to_accum
        elif naccum > 1:
            self.logger.debug('round %d of accumulation', naccum)
            newaccum = self.aggregate_frames(accum, field_to_accum, naccum)
        else:
            msg = 'naccum set to %d, invalid' % (naccum,)
            self.logger.error(msg)
            raise RecipeError(msg)

        # Update partial result
        partial_result.accum = newaccum

        return partial_result

    def aggregate_frames(self, accum, frame, naccum):
        return self.aggregate2(accum, frame, naccum)

    def aggregate2(self, img1, img2, naccum):

        frames = [img1, img2]
        use_errors = True
        # Initial checks
        fframe = frames[0]
        # Ref image
        img = fframe.open()
        has_num_ext = 'NUM' in img
        has_bpm_ext = 'BPM' in img
        base_header = img[0].header
        baseshape = img[0].shape

        data_hdul = []
        for f in frames:
            img = f.open()
            data_hdul.append(img)

        if has_num_ext:
            self.logger.debug('Using NUM extension')
            masks = [numpy.where(m['NUM'].data, 0, 1).astype('uint8') for m in data_hdul]
        elif has_bpm_ext:
            self.logger.debug('Using BPM extension')
            masks = [m['BPM'].data for m in data_hdul]
        else:
            self.logger.warning('BPM missing, use zeros instead')
            false_mask = numpy.zeros(baseshape, dtype='int16')
            masks = [false_mask for _ in data_hdul]

        self.logger.info('Combine target images (final, aggregate)')

        weight_accum = 2 * (1 - 1.0 / naccum)
        weight_frame = 2.0 / naccum
        self.logger.debug("weights for 'accum' and 'frame', %s", [weight_accum, weight_frame])
        scales = [1.0 / weight_accum, 1.0 / weight_frame]
        method = combine.mean
        data_arr = [hdul[0].data for hdul in data_hdul]
        out = method(data_arr, masks=masks, scales=scales, dtype='float32')

        self.logger.debug('create result image')

        return self.create_accum_hdulist(
            data_hdul,
            out,
            method_name=method.__name__,
            use_errors=False
        )




import pytest
import astropy.io.fits as fits
import numpy

import numina.core
import numina.exceptions

from ..subs import BaseABBARecipe

def create_frame_noise(val=0, std=100, pos=3, keys=None):
    size = (100, 100)
    data = numpy.random.normal(val, std, size)
    data[pos] += val
    hdu = fits.PrimaryHDU(data.astype('float32'))
    if keys:
        for k, v in keys.items():
            hdu.header[k] = v

    return fits.HDUList([hdu])


def create_frame(val=0, std=100, pos=3, keys=None):
    size = (100, 100)
    data = numpy.zeros(size)
    data[pos] += val
    hdu = fits.PrimaryHDU(data.astype('float32'))
    if keys:
        for k, v in keys.items():
            hdu.header[k] = v

    return fits.HDUList([hdu])

def create_ob_abba(value, exptime=100.0, starttime=0.0):
    pos = [20, 30, 30, 20]
    frames = []
    nimages = 4
    for i in range(nimages):
        base = starttime
        off1 = i * exptime
        off2 = off1 + exptime
        t1 = base + off1
        t2 = base + off2
        keys = {'TSUTC1': t1, 'TSUTC2': t2, 'NUM-SK': 1}
        frame = numina.core.DataFrame(
            frame=create_frame_noise(
                val=value,
                pos=pos[i],
                keys=keys
            )
        )
        frames.append(frame)

    obsresult = numina.core.ObservationResult()
    obsresult.frames = frames
    obsresult.mode = 'LS_ABBA'
    return obsresult


def test_subs():

    expected_data = numpy.zeros((100, 100))
    expected_data[5] = 4
    expected_data[3] = -4
    keys = {'TSUTC1': 1000000001.00034, 'TSUTC2': 1000000100.00034}
    frames1 = numina.core.DataFrame(frame=create_frame(val=2, pos=5, keys=keys))
    keys = {'TSUTC1': 1000000101.00034, 'TSUTC2': 1000000200.00034}
    frames2 = numina.core.DataFrame(frame=create_frame(val=2, pos=3, keys=keys))
    keys = {'TSUTC1': 1000000201.00034, 'TSUTC2': 1000000300.00034}
    frames3 = numina.core.DataFrame(frame=create_frame(val=2, pos=3, keys=keys))
    keys = {'TSUTC1': 1000000301.00034, 'TSUTC2': 1000000400.00034}
    frames4 = numina.core.DataFrame(frame=create_frame(val=2, pos=5, keys=keys))

    obsresult = numina.core.ObservationResult()
    obsresult.frames = [frames1, frames2, frames3, frames4]

    recipe = BaseABBARecipe()
    rinput = recipe.create_input(
        obresult=obsresult,
    )
    result = recipe.run(rinput)

    reduced_mos_abba_hdul = result.reduced_mos_abba.frame
    assert reduced_mos_abba_hdul[0].header['TSUTC1'] == 1000000001.00034
    assert reduced_mos_abba_hdul[0].header['TSUTC2'] == 1000000400.00034

    assert reduced_mos_abba_hdul[0].header['NUM-NCOM'] == 2
    assert numpy.allclose(expected_data, result.reduced_mos_abba.frame[0].data)


def test_subs_raise():

    expected_data = numpy.zeros((10, 10))
    expected_data[5] = 4
    expected_data[3] = -4
    keys = {'TSUTC1': 1000000001.00034, 'TSUTC2': 1000000100.00034}
    frames1 = numina.core.DataFrame(frame=create_frame(val=2, pos=5, keys=keys))

    obsresult = numina.core.ObservationResult()
    obsresult.frames = [frames1]

    recipe = BaseABBARecipe()
    rinput = BaseABBARecipe.RecipeInput(
        obresult=obsresult,
    )

    with pytest.raises(numina.exceptions.RecipeError):
        recipe.run(rinput)


@pytest.mark.parametrize("nimages", [4])
def test_init_aggregate_result(nimages):
    # Test that works in the first run of the loop
    nstare = 1
    value = 10000
    starttime = 1030040001.00034
    exptime = 105.0

    obsresult = create_ob_abba(value, exptime, starttime)
    obsresult.naccum = 1
    recipe = BaseABBARecipe()
    rinput = recipe.create_input(
        obresult=obsresult,
    )

    #import logging
    #logging.basicConfig(level=logging.DEBUG)
    result1 = recipe.run(rinput)

    frame_hdul = result1.reduced_mos_abba.open()
    accum_hdul = result1.accum.open()

    assert(len(frame_hdul) == len(accum_hdul))

    for key in ['NUM-NCOM', 'TSUTC1', 'TSUTC2']:
        assert frame_hdul[0].header[key] == accum_hdul[0].header[key]

    assert frame_hdul[0].header['NUM-NCOM'] == (nimages / 2) * nstare

    assert numpy.allclose(frame_hdul[0].data, accum_hdul[0].data)


@pytest.mark.parametrize("naccum", [2, 7])
def test_accum_spec(naccum):

    nstare = 1
    nimages = 2
    value = 13.0
    starttime = 1030040001.00034
    exptime = 105.0
    obsresult = create_ob_abba(value, exptime, starttime)
    obsresult.naccum = 1

    recipe = BaseABBARecipe()

    while True:

        rinput = recipe.create_input(
            obresult=obsresult,
        )

        result = recipe.run(rinput)
        frame_hdul = result.reduced_mos_abba.open()
        assert frame_hdul[0].header['NUM-NCOM'] == nimages * nstare
        accum_hdul = result.accum.open()
        # print('frame', obsresult.naccum, frame_hdul[0].data.mean(), frame_hdul[0].data.std(), 2*(1.0/obsresult.naccum))
        # print('acuum', obsresult.naccum, accum_hdul[0].data.mean(), accum_hdul[0].data.std(), 2*(1-1.0/obsresult.naccum))
        assert accum_hdul[0].header['NUM-NCOM'] == nimages * nstare * obsresult.naccum

        if obsresult.naccum < naccum:
            # Init next loop
            nobsresult = create_ob_abba(value, exptime, starttime)
            nobsresult.naccum = obsresult.naccum + 1
            nobsresult.accum = result.accum
            obsresult = nobsresult
        else:
            break

    frame_hdul = result.reduced_mos_abba.open()
    accum_hdul = result.accum.open()

    assert frame_hdul[0].header['NUM-NCOM'] == nimages * nstare
    assert frame_hdul[0].header['TSUTC1'] == starttime
    assert accum_hdul[0].header['NUM-NCOM'] == nimages * nstare * naccum
    assert accum_hdul[0].header['TSUTC1'] == starttime




import pytest

from numina.user.baserun import run_reduce

# Variable for datamanager_remote
TEST_SET_FILE = "pyemir/test/img_acq_set_20200702.tar.gz"


@pytest.mark.result_compare
@pytest.mark.remote_data
def test_img_tut_block1(datamanager_remote):
    task = run_reduce(datamanager_remote, "test1")
    return task.result


@pytest.mark.result_compare
@pytest.mark.remote_data
def test_test_img_tut_block2(datamanager_remote):
    task = run_reduce(datamanager_remote, "test2")
    return task.result


@pytest.mark.remote_data
def test_test_img_tut_block4(datamanager_remote):
    from numina.util.context import working_directory

    with working_directory(datamanager_remote.basedir):
        with open('control_dump.yaml', 'w') as fp:
            datamanager_remote.backend.dump(fp)
        assert True




from numina.core import BaseRecipe

from ..loader import load_drp


def test_recipes_are_defined():

    emir_drp = load_drp()

    assert 'default' in emir_drp.pipelines

    for pipeval in emir_drp.pipelines.values():
        for key, val in pipeval.recipes.items():
            recipe = pipeval.get_recipe_object(key)
            assert isinstance(recipe, BaseRecipe)



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Routines shared by image mode recipes"""

import logging
import operator

import six
import numpy
from astropy.io import fits
import sep
import matplotlib.pyplot as plt

from emirdrp.util.sextractor import SExtractor
from .naming import name_skysub_proc


_logger = logging.getLogger(__name__)

# Actions to carry over images when checking the flux
# of the objects in different images


def warn_action(img):
    _logger.warning('Image %s has low flux in objects', img.baselabel)
    img.valid_science = True


def reject_action(img):
    img.valid_science = False
    _logger.info('Image %s rejected, has low flux in objects', img.baselabel)
    pass


def default_action(img):
    _logger.info(
        'Image %s accepted, has correct flux in objects', img.baselabel)
    img.valid_science = True

# Actions
_dactions = {'warn': warn_action,
             'reject': reject_action, 'default': default_action}


def check_photometry(frames, sf_data, seeing_fwhm, step=0,
                     border=300, extinction=0.0,
                     check_photometry_levels=[0.5, 0.8],
                     check_photometry_actions=['warn', 'warn', 'default'],
                     figure=None):
    # Check photometry of few objects
    weigthmap = 'weights4rms.fits'

    wmap = numpy.ones_like(sf_data[0], dtype='bool')

    # Center of the image
    wmap[border:-border, border:-border] = 0
    # fits.writeto(weigthmap, wmap.astype('uintt8'), overwrite=True)

    basename = 'result_i%0d.fits' % (step)

    data_res = fits.getdata(basename)
    data_res = data_res.byteswap().newbyteorder()
    bkg = sep.Background(data_res)
    data_sub = data_res - bkg

    _logger.info('Runing source extraction tor in %s', basename)
    objects = sep.extract(data_sub, 1.5, err=bkg.globalrms, mask=wmap)

    # if seeing_fwhm is not None:
    #    sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']

    # sex.config['PARAMETERS_LIST'].append('CLASS_STAR')

    # sex.config['CATALOG_NAME'] = 'master-catalogue-i%01d.cat' % step

    LIMIT_AREA = 5000
    idx_small = objects['npix'] < LIMIT_AREA
    objects_small = objects[idx_small]
    NKEEP = 15
    idx_flux = objects_small['flux'].argsort()
    objects_nth = objects_small[idx_flux][-NKEEP:]

    # set of indices of the N first objects
    fluxes = []
    errors = []
    times = []
    airmasses = []

    for idx, frame in enumerate(frames):
        imagename = name_skysub_proc(frame.baselabel, step)

        #sex.config['CATALOG_NAME'] = ('catalogue-%s-i%01d.cat' %
        #                              (frame.baselabel, step))

        # Lauch SExtractor on a FITS file
        # om double image mode
        _logger.info('Runing sextractor in %s', imagename)
        with fits.open(imagename) as hdul:
            header = hdul[0].header
            airmasses.append(header['airmass'])
            times.append(header['tstamp'])
            data_i = hdul[0].data
            data_i = data_i.byteswap().newbyteorder()
            bkg_i = sep.Background(data_i)
            data_sub_i = data_i - bkg_i
        # objects_i = sep.extract(data_sub_i, 1.5, err=bkg_i.globalrms, mask=wmap)
        flux_i, fluxerr_i, flag_i = sep.sum_circle(data_sub_i,
                                                   objects_nth['x'], objects_nth['y'],
                                                   3.0, err=bkg_i.globalrms)


        # Extinction correction
        excor = pow(10, -0.4 * frame.airmass * extinction)

        flux_i = excor * flux_i
        fluxerr_i = excor * fluxerr_i
        fluxes.append(flux_i)
        errors.append(fluxerr_i)

    fluxes_a = numpy.array(fluxes)
    errors_a = numpy.array(errors)
    fluxes_n = fluxes_a / fluxes_a[0]
    errors_a = errors_a / fluxes_a[0]  # sigma
    w = 1.0 / (errors_a) ** 2

    # weighted mean of the flux values
    wdata = numpy.average(fluxes_n, axis=1, weights=w)
    wsigma = 1 / numpy.sqrt(w.sum(axis=1))

    levels = check_photometry_levels
    actions = check_photometry_actions

    x = list(six.moves.range(len(frames)))
    vals, (_, sigma) = check_photometry_categorize(
        x, wdata, levels, tags=actions)
    # n sigma level to plt
    nsig = 3

    if True:
        figure = plt.figure()
        ax = figure.add_subplot(111)

        plot_photometry_check(ax, vals, wsigma, check_photometry_levels, nsig * sigma)
        plt.savefig('figure-relative-flux_i%01d.png' % step)

    for x, _, t in vals:
        try:
            action = _dactions[t]
        except KeyError:
            _logger.warning('Action named %s not recognized, ignoring', t)
            action = default_action
        for p in x:
            action(frames[p])


def check_photometry_categorize(x, y, levels, tags=None):
    '''Put every point in its category.

    levels must be sorted.'''
    x = numpy.asarray(x)
    y = numpy.asarray(y)
    ys = y.copy()
    ys.sort()
    # Mean of the upper half
    m = ys[len(ys) // 2:].mean()
    y /= m
    m = 1.0
    s = ys[len(ys) // 2:].std()
    result = []

    if tags is None:
        tags = list(six.moves.range(len(levels) + 1))

    for l, t in zip(levels, tags):
        indc = y < l
        if indc.any():
            x1 = x[indc]
            y1 = y[indc]
            result.append((x1, y1, t))

            x = x[~indc]
            y = y[~indc]
    else:
        result.append((x, y, tags[-1]))

    return result, (m, s)


def plot_photometry_check(ax, vals, errors, levels, nsigma):
    x = range(len(errors))

    ax.set_title('Relative flux of brightest object')
    for v, c in zip(vals, ['b', 'r', 'g', 'y']):
        ax.scatter(v[0], v[1], c=c)
        w = errors[v[0]]
        ax.errorbar(v[0], v[1], yerr=w, fmt='none', c=c)

    ax.plot([x[0], x[-1]], [1, 1], 'r--')
    ax.plot([x[0], x[-1]], [1 - nsigma, 1 - nsigma], 'b--')
    for f in levels:
        ax.plot([x[0], x[-1]], [f, f], 'g--')

    return ax




def check_position(images_info, sf_data, seeing_fwhm, step=0):
        # FIXME: this method has to be updated

    _logger.info('Checking positions')
    # Check position of bright objects
    weigthmap = 'weights4rms.fits'

    wmap = numpy.zeros_like(sf_data[0])

    # Center of the image
    border = 300
    wmap[border:-border, border:-border] = 1
    fits.writeto(weigthmap, wmap.astype('uint8'), overwrite=True)

    basename = 'result_i%0d.fits' % (step)
    sex = SExtractor()
    sex.config['VERBOSE_TYPE'] = 'QUIET'
    sex.config['PIXEL_SCALE'] = 1
    sex.config['BACK_TYPE'] = 'AUTO'
    if seeing_fwhm is not None and seeing_fwhm > 0:
        sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']
    sex.config['WEIGHT_TYPE'] = 'MAP_WEIGHT'
    sex.config['WEIGHT_IMAGE'] = weigthmap

    sex.config['PARAMETERS_LIST'].append('FLUX_BEST')
    sex.config['PARAMETERS_LIST'].append('FLUXERR_BEST')
    sex.config['PARAMETERS_LIST'].append('FWHM_IMAGE')
    sex.config['PARAMETERS_LIST'].append('CLASS_STAR')

    sex.config['CATALOG_NAME'] = 'master-catalogue-i%01d.cat' % step

    _logger.info('Runing sextractor in %s', basename)
    sex.run('%s,%s' % (basename, basename))

    # Sort catalog by flux
    catalog = sex.catalog()
    catalog = sorted(
        catalog, key=operator.itemgetter('FLUX_BEST'), reverse=True)

    # set of indices of the N first objects
    OBJS_I_KEEP = 10

    # master = [(obj['X_IMAGE'], obj['Y_IMAGE'])
    #    for obj in catalog[:OBJS_I_KEEP]]

    for image in images_info:
        imagename = name_skysub_proc(image.baselabel, step)

        sex.config['CATALOG_NAME'] = ('catalogue-self-%s-i%01d.cat' %
                                      (image.baselabel, step))

        # Lauch SExtractor on a FITS file
        # on double image mode
        _logger.info('Runing sextractor in %s', imagename)
        sex.run(imagename)
        catalog = sex.catalog()

        # data = [(obj['X_IMAGE'], obj['Y_IMAGE']) for obj in catalog]

        # tree = KDTree(data)

        # Search 2 neighbors
        # dists, _ids = tree.query(master, 2, distance_upper_bound=5)

        # for i in dists[:,0]:
        #    print i

        # _logger.info('Mean offset correction for image %s is %f',
        #    imagename, dists[:,0].mean())
        # raw_input('press any key')



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
import numpy as np
from numpy.polynomial.polynomial import polyval
from scipy import ndimage
import sys

from numina.array.display.ximplotxy import ximplotxy
from numina.array.display.ximshow import ximshow
from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.numsplines import AdaptiveLSQUnivariateSpline
from numina.array.wavecalib.apply_integer_offsets import apply_integer_offsets
from numina.tools.arg_file_is_new import arg_file_is_new

from emirdrp.processing.wavecal.slitlet2d import Slitlet2D
from emirdrp.processing.wavecal.set_wv_parameters import set_wv_parameters
from emirdrp.instrument.dtu_configuration import DtuConfiguration
from emirdrp.instrument.csu_configuration import CsuConfiguration
from emirdrp.tools.nscan_minmax_frontiers import nscan_minmax_frontiers
from emirdrp.tools.rect_wpoly_for_mos import islitlet_progress
from emirdrp.tools.save_ndarray_to_fits import save_ndarray_to_fits
from emirdrp.products import RectWaveCoeff

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2
from emirdrp.core import EMIR_NBARS

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: compute pixel-to-pixel flatfield'
    )

    # required arguments
    parser.add_argument("fitsfile",
                        help="Input FITS file (flat ON-OFF)",
                        type=argparse.FileType('rb'))
    parser.add_argument("--rectwv_coeff", required=True,
                        help="Input JSON file with rectification and "
                             "wavelength calibration coefficients",
                        type=argparse.FileType('rt'))
    parser.add_argument("--minimum_slitlet_width_mm", required=True,
                        help="Minimum slitlet width in mm",
                        type=float)
    parser.add_argument("--maximum_slitlet_width_mm", required=True,
                        help="Maximum slitlet width in mm",
                        type=float)
    parser.add_argument("--minimum_fraction", required=True,
                        help="Minimum allowed flatfielding value",
                        type=float, default=0.01)
    parser.add_argument("--minimum_value_in_output",
                        help="Minimum value allowed in output file: pixels "
                             "below this value are set to 1.0 (default=0.01)",
                        type=float, default=0.01)
    parser.add_argument("--maximum_value_in_output",
                        help="Maximum value allowed in output file: pixels "
                             "above this value are set to 1.0 (default=10.0)",
                        type=float, default=10.0)
    parser.add_argument("--nwindow_median",
                        help="Window size to smooth median spectrum in the "
                             "spectral direction",
                        type=int)
    parser.add_argument("--outfile", required=True,
                        help="Output FITS file",
                        type=lambda x: arg_file_is_new(parser, x, mode='wb'))

    # optional arguments
    parser.add_argument("--delta_global_integer_offset_x_pix",
                        help="Delta global integer offset in the X direction "
                             "(default=0)",
                        default=0, type=int)
    parser.add_argument("--delta_global_integer_offset_y_pix",
                        help="Delta global integer offset in the Y direction "
                             "(default=0)",
                        default=0, type=int)
    parser.add_argument("--resampling",
                        help="Resampling method: 1 -> nearest neighbor, "
                             "2 -> linear interpolation (default)",
                        default=2, type=int,
                        choices=(1, 2))
    parser.add_argument("--ignore_DTUconf",
                        help="Ignore DTU configurations differences between "
                             "model and input image",
                        action="store_true")
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting & debugging options"
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # This code is obsolete
    raise ValueError('This code is obsolete: use recipe in '
                     'emirdrp/recipes/spec/flatpix2pix.py')

    # read calibration structure from JSON file
    rectwv_coeff = RectWaveCoeff._datatype_load(args.rectwv_coeff.name)

    # modify (when requested) global offsets
    rectwv_coeff.global_integer_offset_x_pix += \
        args.delta_global_integer_offset_x_pix
    rectwv_coeff.global_integer_offset_y_pix += \
        args.delta_global_integer_offset_y_pix

    # read FITS image and its corresponding header
    hdulist = fits.open(args.fitsfile)
    header = hdulist[0].header
    image2d = hdulist[0].data
    hdulist.close()

    # apply global offsets
    image2d = apply_integer_offsets(
        image2d=image2d,
        offx=rectwv_coeff.global_integer_offset_x_pix,
        offy=rectwv_coeff.global_integer_offset_y_pix
    )

    # protections
    naxis2, naxis1 = image2d.shape
    if naxis1 != header['naxis1'] or naxis2 != header['naxis2']:
        print('>>> NAXIS1:', naxis1)
        print('>>> NAXIS2:', naxis2)
        raise ValueError('Something is wrong with NAXIS1 and/or NAXIS2')
    if abs(args.debugplot) >= 10:
        print('>>> NAXIS1:', naxis1)
        print('>>> NAXIS2:', naxis2)

    # check that the input FITS file grism and filter match
    filter_name = header['filter']
    if filter_name != rectwv_coeff.tags['filter']:
        raise ValueError("Filter name does not match!")
    grism_name = header['grism']
    if grism_name != rectwv_coeff.tags['grism']:
        raise ValueError("Filter name does not match!")
    if abs(args.debugplot) >= 10:
        print('>>> grism.......:', grism_name)
        print('>>> filter......:', filter_name)

    # check that the DTU configurations are compatible
    dtu_conf_fitsfile = DtuConfiguration.define_from_fits(args.fitsfile)
    dtu_conf_jsonfile = DtuConfiguration.define_from_dictionary(
        rectwv_coeff.meta_info['dtu_configuration'])
    if dtu_conf_fitsfile != dtu_conf_jsonfile:
        print('DTU configuration (FITS file):\n\t', dtu_conf_fitsfile)
        print('DTU configuration (JSON file):\n\t', dtu_conf_jsonfile)
        if args.ignore_DTUconf:
            print('WARNING: DTU configuration differences found!')
        else:
            raise ValueError('DTU configurations do not match')
    else:
        if abs(args.debugplot) >= 10:
            print('>>> DTU Configuration match!')
            print(dtu_conf_fitsfile)

    # load CSU configuration
    csu_conf_fitsfile = CsuConfiguration.define_from_fits(args.fitsfile)
    if abs(args.debugplot) >= 10:
        print(csu_conf_fitsfile)

    # valid slitlet numbers
    list_valid_islitlets = list(range(1, EMIR_NBARS + 1))
    for idel in rectwv_coeff.missing_slitlets:
        print('-> Removing slitlet (not defined):', idel)
        list_valid_islitlets.remove(idel)
    # filter out slitlets with widths outside valid range
    list_outside_valid_width = []
    for islitlet in list_valid_islitlets:
        slitwidth = csu_conf_fitsfile.csu_bar_slit_width(islitlet)
        if (slitwidth < args.minimum_slitlet_width_mm) or \
                (slitwidth > args.maximum_slitlet_width_mm):
            list_outside_valid_width.append(islitlet)
            print('-> Removing slitlet (invalid width):', islitlet)
    if len(list_outside_valid_width) > 0:
        for idel in list_outside_valid_width:
            list_valid_islitlets.remove(idel)
    print('>>> valid slitlet numbers:\n', list_valid_islitlets)

    # ---

    # compute and store median spectrum (and masked region) for each
    # individual slitlet
    image2d_sp_median = np.zeros((EMIR_NBARS, EMIR_NAXIS1))
    image2d_sp_mask = np.zeros((EMIR_NBARS, EMIR_NAXIS1), dtype=bool)
    for islitlet in list(range(1, EMIR_NBARS + 1)):
        if islitlet in list_valid_islitlets:
            if args.debugplot == 0:
                islitlet_progress(islitlet, EMIR_NBARS, ignore=False)
            # define Slitlet2D object
            slt = Slitlet2D(islitlet=islitlet,
                            rectwv_coeff=rectwv_coeff,
                            debugplot=args.debugplot)

            if abs(args.debugplot) >= 10:
                print(slt)

            # extract (distorted) slitlet from the initial image
            slitlet2d = slt.extract_slitlet2d(
                image_2k2k=image2d,
                subtitle='original image'
            )

            # rectify slitlet
            slitlet2d_rect = slt.rectify(
                slitlet2d=slitlet2d,
                resampling=args.resampling,
                subtitle='original rectified'
            )
            naxis2_slitlet2d, naxis1_slitlet2d = slitlet2d_rect.shape

            if naxis1_slitlet2d != EMIR_NAXIS1:
                print('naxis1_slitlet2d: ', naxis1_slitlet2d)
                print('EMIR_NAXIS1.....: ', EMIR_NAXIS1)
                raise ValueError("Unexpected naxis1_slitlet2d")

            sp_mask = np.zeros(naxis1_slitlet2d, dtype=bool)

            # for grism LR set to zero data beyond useful wavelength range
            if grism_name == 'LR':
                wv_parameters = set_wv_parameters(filter_name, grism_name)
                x_pix = np.arange(1, naxis1_slitlet2d + 1)
                wl_pix = polyval(x_pix, slt.wpoly)
                lremove = wl_pix < wv_parameters['wvmin_useful']
                sp_mask[lremove] = True
                slitlet2d_rect[:, lremove] = 0.0
                lremove = wl_pix > wv_parameters['wvmax_useful']
                slitlet2d_rect[:, lremove] = 0.0
                sp_mask[lremove] = True

            # get useful slitlet region (use boundaries instead of frontiers;
            # note that the nscan_minmax_frontiers() works well independently
            # of using frontiers of boundaries as arguments)
            nscan_min, nscan_max = nscan_minmax_frontiers(
                slt.y0_reference_lower,
                slt.y0_reference_upper,
                resize=False
            )
            ii1 = nscan_min - slt.bb_ns1_orig
            ii2 = nscan_max - slt.bb_ns1_orig + 1

            # median spectrum
            sp_collapsed = np.median(slitlet2d_rect[ii1:(ii2 + 1), :], axis=0)

            # smooth median spectrum along the spectral direction
            sp_median = ndimage.median_filter(
                sp_collapsed,
                args.nwindow_median,
                mode='nearest'
            )

            """
                nremove = 5
                spl = AdaptiveLSQUnivariateSpline(
                    x=xaxis1[nremove:-nremove],
                    y=sp_collapsed[nremove:-nremove],
                    t=11,
                    adaptive=True
                )
                xknots = spl.get_knots()
                yknots = spl(xknots)
                sp_median = spl(xaxis1)

                # compute rms within each knot interval
                nknots = len(xknots)
                rms_array = np.zeros(nknots - 1, dtype=float)
                for iknot in range(nknots - 1):
                    residuals = []
                    for xdum, ydum, yydum in \
                            zip(xaxis1, sp_collapsed, sp_median):
                        if xknots[iknot] <= xdum <= xknots[iknot + 1]:
                            residuals.append(abs(ydum - yydum))
                    if len(residuals) > 5:
                        rms_array[iknot] = np.std(residuals)
                    else:
                        rms_array[iknot] = 0

                # determine in which knot interval falls each pixel
                iknot_array = np.zeros(len(xaxis1), dtype=int)
                for idum, xdum in enumerate(xaxis1):
                    for iknot in range(nknots - 1):
                        if xknots[iknot] <= xdum <= xknots[iknot + 1]:
                            iknot_array[idum] = iknot

                # compute new fit removing deviant points (with fixed knots)
                xnewfit = []
                ynewfit = []
                for idum in range(len(xaxis1)):
                    delta_sp = abs(sp_collapsed[idum] - sp_median[idum])
                    rms_tmp = rms_array[iknot_array[idum]]
                    if idum == 0 or idum == (len(xaxis1) - 1):
                        lok = True
                    elif rms_tmp > 0:
                        if delta_sp < 3.0 * rms_tmp:
                            lok = True
                        else:
                            lok = False
                    else:
                        lok = True
                    if lok:
                        xnewfit.append(xaxis1[idum])
                        ynewfit.append(sp_collapsed[idum])
                nremove = 5
                splnew = AdaptiveLSQUnivariateSpline(
                    x=xnewfit[nremove:-nremove],
                    y=ynewfit[nremove:-nremove],
                    t=xknots[1:-1],
                    adaptive=False
                )
                sp_median = splnew(xaxis1)
            """

            ymax_spmedian = sp_median.max()
            y_threshold = ymax_spmedian * args.minimum_fraction
            lremove = np.where(sp_median < y_threshold)
            sp_median[lremove] = 0.0
            sp_mask[lremove] = True

            image2d_sp_median[islitlet - 1, :] = sp_median
            image2d_sp_mask[islitlet - 1, :] = sp_mask

            if abs(args.debugplot) % 10 != 0:
                xaxis1 = np.arange(1, naxis1_slitlet2d + 1)
                title = 'Slitlet#' + str(islitlet) + ' (median spectrum)'
                ax = ximplotxy(xaxis1, sp_collapsed,
                               title=title,
                               show=False, **{'label' : 'collapsed spectrum'})
                ax.plot(xaxis1, sp_median, label='fitted spectrum')
                ax.plot([1, naxis1_slitlet2d], 2*[y_threshold],
                        label='threshold')
                # ax.plot(xknots, yknots, 'o', label='knots')
                ax.legend()
                ax.set_ylim(-0.05*ymax_spmedian, 1.05*ymax_spmedian)
                pause_debugplot(args.debugplot,
                                pltshow=True, tight_layout=True)
        else:
            if args.debugplot == 0:
                islitlet_progress(islitlet, EMIR_NBARS, ignore=True)

    # ToDo: compute "average" spectrum for each pseudo-longslit, scaling
    #       with the median signal in each slitlet; derive a particular
    #       spectrum for each slitlet (scaling properly)

    image2d_sp_median_masked = np.ma.masked_array(
        image2d_sp_median,
        mask=image2d_sp_mask
    )
    ycut_median = np.ma.median(image2d_sp_median_masked, axis=1).data
    ycut_median_2d = np.repeat(ycut_median, EMIR_NAXIS1).reshape(
        EMIR_NBARS, EMIR_NAXIS1)
    image2d_sp_median_eq = image2d_sp_median_masked / ycut_median_2d
    image2d_sp_median_eq = image2d_sp_median_eq.data

    if True:
        ximshow(image2d_sp_median, title='sp_median', debugplot=12)
        ximplotxy(np.arange(1, EMIR_NBARS + 1), ycut_median, 'ro',
                  title='median value of each spectrum', debugplot=12)
        ximshow(image2d_sp_median_eq, title='sp_median_eq', debugplot=12)

    csu_conf_fitsfile.display_pseudo_longslits(
        list_valid_slitlets=list_valid_islitlets)
    dict_longslits = csu_conf_fitsfile.pseudo_longslits()

    # compute median spectrum for each longslit and insert (properly
    # scaled) that spectrum in each slitlet belonging to that longslit
    image2d_sp_median_longslit = np.zeros((EMIR_NBARS, EMIR_NAXIS1))
    islitlet = 1
    loop = True
    while loop:
        if islitlet in list_valid_islitlets:
            imin = dict_longslits[islitlet].imin()
            imax = dict_longslits[islitlet].imax()
            print('--> imin, imax: ', imin, imax)
            sp_median_longslit = np.median(
                image2d_sp_median_eq[(imin - 1):imax, :], axis=0)
            for i in range(imin, imax+1):
                print('----> i: ', i)
                image2d_sp_median_longslit[(i - 1), :] = \
                    sp_median_longslit * ycut_median[i - 1]
            islitlet = imax
        else:
            print('--> ignoring: ', islitlet)
        if islitlet == EMIR_NBARS:
            loop = False
        else:
            islitlet += 1
    if True:
        ximshow(image2d_sp_median_longslit, debugplot=12)

    # initialize rectified image
    image2d_flatfielded = np.zeros((EMIR_NAXIS2, EMIR_NAXIS1))

    # main loop
    for islitlet in list(range(1, EMIR_NBARS + 1)):
        if islitlet in list_valid_islitlets:
            if args.debugplot == 0:
                islitlet_progress(islitlet, EMIR_NBARS, ignore=False)
            # define Slitlet2D object
            slt = Slitlet2D(islitlet=islitlet,
                            rectwv_coeff=rectwv_coeff,
                            debugplot=args.debugplot)

            # extract (distorted) slitlet from the initial image
            slitlet2d = slt.extract_slitlet2d(
                image_2k2k=image2d,
                subtitle='original image'
            )

            # rectify slitlet
            slitlet2d_rect = slt.rectify(
                slitlet2d=slitlet2d,
                resampling=args.resampling,
                subtitle='original rectified'
            )
            naxis2_slitlet2d, naxis1_slitlet2d = slitlet2d_rect.shape

            sp_median = image2d_sp_median_longslit[islitlet - 1, :]

            # generate rectified slitlet region filled with the median spectrum
            slitlet2d_rect_spmedian = np.tile(sp_median, (naxis2_slitlet2d, 1))
            if abs(args.debugplot) > 10:
                slt.ximshow_rectified(
                    slitlet2d_rect=slitlet2d_rect_spmedian,
                    subtitle='rectified, filled with median spectrum'
                )

            # unrectified image
            slitlet2d_unrect_spmedian = slt.rectify(
                slitlet2d=slitlet2d_rect_spmedian,
                resampling=args.resampling,
                inverse=True,
                subtitle='unrectified, filled with median spectrum'
            )

            # normalize initial slitlet image (avoid division by zero)
            slitlet2d_norm = np.zeros_like(slitlet2d)
            for j in range(naxis1_slitlet2d):
                for i in range(naxis2_slitlet2d):
                    den = slitlet2d_unrect_spmedian[i, j]
                    if den == 0:
                        slitlet2d_norm[i, j] = 1.0
                    else:
                        slitlet2d_norm[i, j] = slitlet2d[i, j] / den

            if abs(args.debugplot) > 10:
                slt.ximshow_unrectified(
                    slitlet2d=slitlet2d_norm,
                    subtitle='unrectified, pixel-to-pixel'
                )

            # check for pseudo-longslit with previous slitlet
            if islitlet > 1:
                if (islitlet - 1) in list_valid_islitlets:
                    c1 = csu_conf_fitsfile.csu_bar_slit_center(islitlet - 1)
                    w1 = csu_conf_fitsfile.csu_bar_slit_width(islitlet - 1)
                    c2 = csu_conf_fitsfile.csu_bar_slit_center(islitlet)
                    w2 = csu_conf_fitsfile.csu_bar_slit_width(islitlet)
                    if abs(w1-w2)/w1 < 0.25:
                        wmean = (w1 + w2) / 2.0
                        if abs(c1 - c2) < wmean/4.0:
                            same_slitlet_below = True
                        else:
                            same_slitlet_below = False
                    else:
                        same_slitlet_below = False
                else:
                    same_slitlet_below = False
            else:
                same_slitlet_below = False

            # check for pseudo-longslit with next slitlet
            if islitlet < EMIR_NBARS:
                if (islitlet + 1) in list_valid_islitlets:
                    c1 = csu_conf_fitsfile.csu_bar_slit_center(islitlet)
                    w1 = csu_conf_fitsfile.csu_bar_slit_width(islitlet)
                    c2 = csu_conf_fitsfile.csu_bar_slit_center(islitlet + 1)
                    w2 = csu_conf_fitsfile.csu_bar_slit_width(islitlet + 1)
                    if abs(w1-w2)/w1 < 0.25:
                        wmean = (w1 + w2) / 2.0
                        if abs(c1 - c2) < wmean/4.0:
                            same_slitlet_above = True
                        else:
                            same_slitlet_above = False
                    else:
                        same_slitlet_above = False
                else:
                    same_slitlet_above = False
            else:
                same_slitlet_above = False

            for j in range(EMIR_NAXIS1):
                xchannel = j + 1
                y0_lower = slt.list_frontiers[0](xchannel)
                y0_upper = slt.list_frontiers[1](xchannel)
                n1, n2 = nscan_minmax_frontiers(y0_frontier_lower=y0_lower,
                                                y0_frontier_upper=y0_upper,
                                                resize=True)
                # note that n1 and n2 are scans (ranging from 1 to NAXIS2)
                nn1 = n1 - slt.bb_ns1_orig + 1
                nn2 = n2 - slt.bb_ns1_orig + 1
                image2d_flatfielded[(n1 - 1):n2, j] = \
                    slitlet2d_norm[(nn1 - 1):nn2, j]

                # force to 1.0 region around frontiers
                if not same_slitlet_below:
                    image2d_flatfielded[(n1 - 1):(n1 + 2), j] = 1
                if not same_slitlet_above:
                    image2d_flatfielded[(n2 - 5):n2, j] = 1
        else:
            if args.debugplot == 0:
                islitlet_progress(islitlet, EMIR_NBARS, ignore=True)

    if args.debugplot == 0:
        print('OK!')

    # restore global offsets
    image2d_flatfielded = apply_integer_offsets(
        image2d=image2d_flatfielded ,
        offx=-rectwv_coeff.global_integer_offset_x_pix,
        offy=-rectwv_coeff.global_integer_offset_y_pix
    )

    # set pixels below minimum value to 1.0
    filtered = np.where(image2d_flatfielded < args.minimum_value_in_output)
    image2d_flatfielded[filtered] = 1.0

    # set pixels above maximum value to 1.0
    filtered = np.where(image2d_flatfielded > args.maximum_value_in_output)
    image2d_flatfielded[filtered] = 1.0

    # save output file
    save_ndarray_to_fits(
        array=image2d_flatfielded,
        file_name=args.outfile,
        main_header=header,
        overwrite=True
    )
    print('>>> Saving file ' + args.outfile.name)


if __name__ == "__main__":
    main()



#
# Copyright 2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""EMIR constants with units"""

import astropy.units as u

# FIXME: duplicated in emirdrp.core

EMIR_PIXSIZE = 18.0 * u.micron

EMIR_PIXSCALE = 0.1944 * u.arcsec / u.pixel

# Platescale in focal plane of Nasmyth A
GTC_NASMYTH_A_PLATESCALE = 1.212 * u.arcsec / u.mm

# Plate scale, focal plane in the window of the EMIR
# slightly different from NASMYTH_A_PLATESCALE
# keyword PSCFP
EMIR_CSU_PLATESCALE = 1.1746 * u.arcsec / u.mm

# plate scale, detector, arcsec / mm
EMIR_DETECTOR_PLATESCALE = 10.8  * u.arcsec / u.mm

# This value is stored in keyword IPA
EMIR_REF_IPA = 90.0552 * u.deg # deg



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from __future__ import division
from __future__ import print_function

from copy import deepcopy
import logging
import numpy as np

from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximplotxy import ximplotxy
from numina.array.display.matplotlib_qt import plt
from numina.array.display.matplotlib_qt import set_window_geometry
from numina.array.stats import summary
from numina.array.wavecalib.check_wlcalib import check_wlcalib_sp
from numina.array.wavecalib.crosscorrelation import convolve_comb_lines
from numina.array.wavecalib.crosscorrelation import periodic_corr1d
from numina.array.wavecalib.fix_pix_borders import find_pix_borders
from numina.frame.utils import copy_img

import emirdrp.datamodel as datamodel
from emirdrp.instrument.csu_configuration import CsuConfiguration
from emirdrp.processing.wavecal.median_slitlets_rectified \
    import median_slitlets_rectified
from emirdrp.processing.wavecal.set_wv_parameters import set_wv_parameters

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NBARS


def refine_rectwv_coeff(input_image, rectwv_coeff,
                        catlines_all_wave,
                        catlines_all_flux,
                        refine_wavecalib_mode,
                        list_useful_slitlets,
                        save_intermediate_results=False,
                        debugplot=0):
    """Refine RectWaveCoeff object using a catalogue of lines

    One and only one among refine_with_oh_lines_mode and
    refine_with_arc_lines must be different from zero.

    Parameters
    ----------
    input_image : HDUList object
        Input 2D image.
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for the
        particular CSU configuration.
    catlines_all_wave : numpy array
        Array with wavelengths
    catlines_all_flux : numpy array
        Array with fluxes.
    refine_wavecalib_mode : int
        Integer, indicating the type of refinement:
        0 : no refinement
        1 : apply the same global offset to all the slitlets (using ARC lines)
        2 : apply individual offset to each slitlet (using ARC lines)
        11 : apply the same global offset to all the slitlets (using OH lines)
        12 : apply individual offset to each slitlet (using OH lines)
    list_useful_slitlets : list of integers
        List of useful slitlets.
    save_intermediate_results : bool
        If True, save plots in PDF files
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    refined_rectwv_coeff : RectWaveCoeff instance
        Refined rectification and wavelength calibration coefficients
        for the particular CSU configuration.
    expected_cat_image : HDUList object
        Output 2D image (rectified and wavelength calibrated) with
        the expected catalogue lines.

    """

    logger = logging.getLogger(__name__)

    if save_intermediate_results:
        from matplotlib.backends.backend_pdf import PdfPages
        pdf = PdfPages('crosscorrelation.pdf')
    else:
        pdf = None

    # image header
    main_header = input_image[0].header
    mecs_header = datamodel.get_mecs_header(input_image)
    filter_name = main_header['filter']
    grism_name = main_header['grism']

    # initialize output
    refined_rectwv_coeff = deepcopy(rectwv_coeff)

    logger.info('Computing median spectrum')
    # compute median spectrum and normalize it
    sp_median = median_slitlets_rectified(
        input_image,
        mode=2,
        list_useful_slitlets=list_useful_slitlets
    )[0].data
    sp_median /= sp_median.max()

    # determine minimum and maximum useful wavelength
    jmin, jmax = find_pix_borders(sp_median, 0)
    naxis1 = main_header['naxis1']
    naxis2 = main_header['naxis2']
    crpix1 = main_header['crpix1']
    crval1 = main_header['crval1']
    cdelt1 = main_header['cdelt1']
    xwave = crval1 + (np.arange(naxis1) + 1.0 - crpix1) * cdelt1
    if grism_name == 'LR':
        wv_parameters = set_wv_parameters(filter_name, grism_name)
        wave_min = wv_parameters['wvmin_useful']
        wave_max = wv_parameters['wvmax_useful']
    else:
        wave_min = crval1 + (jmin + 1 - crpix1) * cdelt1
        wave_max = crval1 + (jmax + 1 - crpix1) * cdelt1
    logger.info('Setting wave_min to {}'.format(wave_min))
    logger.info('Setting wave_max to {}'.format(wave_max))

    # extract subset of catalogue lines within current wavelength range
    lok1 = catlines_all_wave >= wave_min
    lok2 = catlines_all_wave <= wave_max
    catlines_reference_wave = catlines_all_wave[lok1*lok2]
    catlines_reference_flux = catlines_all_flux[lok1*lok2]
    catlines_reference_flux /= catlines_reference_flux.max()

    # estimate sigma to broaden catalogue lines
    csu_config = CsuConfiguration.define_from_header(mecs_header)
    # segregate slitlets
    list_not_useful_slitlets = [i for i in list(range(1, EMIR_NBARS + 1))
                                if i not in list_useful_slitlets]
    logger.info('list of useful slitlets: {}'.format(
        list_useful_slitlets))
    logger.info('list of unusable slitlets: {}'.format(
        list_not_useful_slitlets))
    tempwidths = np.array([csu_config.csu_bar_slit_width(islitlet)
                           for islitlet in list_useful_slitlets])
    widths_summary = summary(tempwidths)
    logger.info('Statistics of useful slitlet widths (mm):')
    logger.info('- npoints....: {0:d}'.format(widths_summary['npoints']))
    logger.info('- mean.......: {0:7.3f}'.format(widths_summary['mean']))
    logger.info('- median.....: {0:7.3f}'.format(widths_summary['median']))
    logger.info('- std........: {0:7.3f}'.format(widths_summary['std']))
    logger.info('- robust_std.: {0:7.3f}'.format(widths_summary['robust_std']))
    # empirical transformation of slit width (mm) to pixels
    sigma_broadening = cdelt1 * widths_summary['median']

    # convolve location of catalogue lines to generate expected spectrum
    xwave_reference, sp_reference = convolve_comb_lines(
        catlines_reference_wave, catlines_reference_flux, sigma_broadening,
        crpix1, crval1, cdelt1, naxis1
    )
    sp_reference /= sp_reference.max()

    # generate image2d with expected lines
    if save_intermediate_results:
        image2d_expected_lines = np.tile(sp_reference, (naxis2, 1))
        expected_cat_image = copy_img(input_image)
        expected_cat_image[0].data = image2d_expected_lines
    else:
        expected_cat_image = None

    if (abs(debugplot) % 10 != 0) or (pdf is not None):
        ax = ximplotxy(xwave, sp_median, 'C1-',
                       xlabel='Wavelength (Angstroms, in vacuum)',
                       ylabel='Normalized number of counts',
                       title='Median spectrum',
                       label='observed spectrum', show=False)
        # overplot reference catalogue lines
        ax.stem(catlines_reference_wave, catlines_reference_flux, 'C4-',
                markerfmt=' ', basefmt='C4-', label='tabulated lines')
        # overplot convolved reference lines
        ax.plot(xwave_reference, sp_reference, 'C0-',
                label='expected spectrum')
        ax.legend()
        if pdf is not None:
            pdf.savefig()
        else:
            pause_debugplot(debugplot=debugplot, pltshow=True)

    # compute baseline signal in sp_median
    baseline = np.percentile(sp_median[sp_median > 0], q=10)
    if (abs(debugplot) % 10 != 0) or (pdf is not None):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.hist(sp_median, bins=1000, log=True)
        ax.set_xlabel('Normalized number of counts')
        ax.set_ylabel('Number of pixels')
        ax.set_title('Median spectrum')
        ax.axvline(float(baseline), linestyle='--', color='grey')
        if pdf is not None:
            pdf.savefig()
        else:
            geometry = (0, 0, 640, 480)
            set_window_geometry(geometry)
            plt.show()
    # subtract baseline to sp_median (only pixels with signal above zero)
    lok = np.where(sp_median > 0)
    sp_median[lok] -= baseline

    # compute global offset through periodic correlation
    logger.info('Computing global offset')
    global_offset, fpeak = periodic_corr1d(
        sp_reference=sp_reference,
        sp_offset=sp_median,
        fminmax=None,
        naround_zero=50,
        plottitle='Median spectrum (cross-correlation)',
        pdf=pdf,
        debugplot=debugplot
    )
    logger.info('Global offset: {} pixels'.format(-global_offset))

    missing_slitlets = rectwv_coeff.missing_slitlets

    if refine_wavecalib_mode in [1, 11]:
        # apply computed offset to obtain refined_rectwv_coeff_global
        for islitlet in range(1, EMIR_NBARS + 1):
            if islitlet not in missing_slitlets:
                i = islitlet - 1
                dumdict = refined_rectwv_coeff.contents[i]
                dumdict['wpoly_coeff'][0] -= global_offset*cdelt1

    elif refine_wavecalib_mode in [2, 12]:
        # compute individual offset for each slitlet
        logger.info('Computing individual offsets')
        median_55sp = median_slitlets_rectified(input_image, mode=1)
        offset_array = np.zeros(EMIR_NBARS)
        xplot = []
        yplot = []
        xplot_skipped = []
        yplot_skipped = []
        cout = '0'
        for islitlet in range(1, EMIR_NBARS + 1):
            if islitlet in list_useful_slitlets:
                i = islitlet - 1
                sp_median = median_55sp[0].data[i, :]
                lok = np.where(sp_median > 0)
                if np.any(lok):
                    baseline = np.percentile(sp_median[lok], q=10)
                    sp_median[lok] -= baseline
                    sp_median /= sp_median.max()
                    offset_array[i], fpeak = periodic_corr1d(
                        sp_reference=sp_reference,
                        sp_offset=median_55sp[0].data[i, :],
                        fminmax=None,
                        naround_zero=50,
                        plottitle='slitlet #{0} (cross-correlation)'.format(
                            islitlet),
                        pdf=pdf,
                        debugplot=debugplot
                    )
                else:
                    offset_array[i] = 0.0
                dumdict = refined_rectwv_coeff.contents[i]
                dumdict['wpoly_coeff'][0] -= offset_array[i]*cdelt1
                xplot.append(islitlet)
                yplot.append(-offset_array[i])
                # second correction
                wpoly_coeff_refined = check_wlcalib_sp(
                    sp=median_55sp[0].data[i, :],
                    crpix1=crpix1,
                    crval1=crval1-offset_array[i]*cdelt1,
                    cdelt1=cdelt1,
                    wv_master=catlines_reference_wave,
                    coeff_ini=dumdict['wpoly_coeff'],
                    naxis1_ini=EMIR_NAXIS1,
                    title='slitlet #{0} (after applying offset)'.format(
                        islitlet),
                    ylogscale=False,
                    pdf=pdf,
                    debugplot=debugplot
                )
                dumdict['wpoly_coeff'] = wpoly_coeff_refined
                cout += '.'

            else:
                xplot_skipped.append(islitlet)
                yplot_skipped.append(0)
                cout += 'i'

            if islitlet % 10 == 0:
                if cout != 'i':
                    cout = str(islitlet // 10)

            logger.info(cout)

        # show offsets with opposite sign
        stat_summary = summary(np.array(yplot))
        logger.info('Statistics of individual slitlet offsets (pixels):')
        logger.info('- npoints....: {0:d}'.format(stat_summary['npoints']))
        logger.info('- mean.......: {0:7.3f}'.format(stat_summary['mean']))
        logger.info('- median.....: {0:7.3f}'.format(stat_summary['median']))
        logger.info('- std........: {0:7.3f}'.format(stat_summary['std']))
        logger.info('- robust_std.: {0:7.3f}'.format(stat_summary[
                                                        'robust_std']))
        if (abs(debugplot) % 10 != 0) or (pdf is not None):
            ax = ximplotxy(xplot, yplot,
                           linestyle='', marker='o', color='C0',
                           xlabel='slitlet number',
                           ylabel='-offset (pixels) = offset to be applied',
                           title='cross-correlation result',
                           show=False, **{'label': 'individual slitlets'})
            if len(xplot_skipped) > 0:
                ax.plot(xplot_skipped, yplot_skipped, 'mx')
            ax.axhline(-global_offset, linestyle='--', color='C1',
                       label='global offset')
            ax.legend()
            if pdf is not None:
                pdf.savefig()
            else:
                pause_debugplot(debugplot=debugplot, pltshow=True)
    else:
        raise ValueError('Unexpected mode={}'.format(refine_wavecalib_mode))

    # close output PDF file
    if pdf is not None:
        pdf.close()

    # return result
    return refined_rectwv_coeff, expected_cat_image



#
# Copyright 2008-2015 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

def braid(*iterables):
    """Return the elements of each iterator in turn until some is exhausted.
    This function is similar to the roundrobin example
    in itertools documentation.
    >>> a = iter([1,2,3,4])
    >>> b = iter(['a', 'b'])
    >>> c = iter([1,1,1,1,'a', 'c'])
    >>> d = iter([1,1,1,1,1,1])
    >>> list(braid(a, b, c, d))
    [1, 'a', 1, 1, 2, 'b', 1, 1]
    """

    for itbl in zip(*iterables):
        for it in itbl:
            yield it


_P1 = slice(0, 1024)
_P2 = slice(1024, 2048)
_O1 = [slice(i*128,(i+1)*128) for i in range(8)]
_O2 = [slice(1024 + (7-i)*128, 1024+(8-i)*128) for i in range(8)]


_CH1 = [(_P2, s) for s in _O2]
_CH2 = [(s, _P1) for s in _O2]
_CH3 = [(_P1, s) for s in _O1]
_CH4 = [(s, _P2) for s in _O1]

# Channels are listed per quadrant and then in fast readout order
CHANNELS = [chan for chans in [_CH1, _CH2, _CH3, _CH4] for chan in chans]
CHANNELS_1 = CHANNELS
# Channels as they are populated during reconstruction
CHANNELS_2 = [chan for chans in [_CH3, _CH4, _CH1, _CH2] for chan in chans]
# Channels as listed in Carlos Gonzalez Ph. D. Thesis
CHANNELS_3 = [chan for chans in [_CH2, _CH3, _CH4, _CH1] for chan in reversed(chans)]
# Channels in readout order
CHANNELS_READOUT = list(braid(_CH3,_CH4, _CH1, _CH2))

RCHANNELS_1 = [chan for chans in [_CH3, _CH1, _CH2, _CH4] for chan in chans]
FULL = RCHANNELS_1


# Quadrants are listed starting at left-top and counter-clockwise then
QUADRANTS = [(_P2, _P1), (_P1, _P1), (_P1, _P2), (_P2, _P2)]


# FIXME: this is a hack to convert channel name to a structure
def convert_name_to_channels(conf):
    chname = conf.configuration['detector']['channels']
    allcha = globals()[chname]
    conf.configuration['detector']['channels'] = allcha
    return conf



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
import numpy as np
import sys

from numina.array.wavecalib.fix_pix_borders import find_pix_borders
from numina.tools.arg_file_is_new import arg_file_is_new

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES


def merge2images(hdu1, hdu2, debugplot):
    """Merge 2 EMIR images, averaging the common region.

    Parameters
    ----------
    hdu1 : HDUList object
        Input image #1.
    hdu2 : HDUList object
        Input image #2.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    image_merged : HDUList object
        Output image.

    """

    # check image dimensions
    image_header = hdu1[0].header
    image_header_ = hdu2[0].header
    #
    naxis = image_header['naxis']
    naxis_ = image_header_['naxis']
    if naxis != naxis_:
        raise ValueError('Incompatible NAXIS values: {}, {}'.format(
            naxis, naxis_
        ))
    naxis1 = image_header['naxis1']
    naxis1_ = image_header_['naxis1']
    if naxis1 != naxis1_:
        raise ValueError('Incompatible NAXIS1 values: {}, {}'.format(
            naxis1, naxis1_
        ))
    if naxis == 1:
        naxis2 = 1
    elif naxis == 2:
        naxis2 = image_header['naxis2']
        naxis2_ = image_header_['naxis2']
        if naxis2 != naxis2_:
            raise ValueError('Incompatible NAXIS2 values: {}, {}'.format(
                naxis2, naxis2_
            ))
    else:
        raise ValueError('Unexpected NAXIS value: {}'.format(naxis))

    # initialize output array
    image2d_merged = np.zeros((naxis2, naxis1))

    # main loop
    data1 = hdu1[0].data
    data2 = hdu2[0].data
    for i in range(naxis2):
        sp1 = data1[i, :]
        sp2 = data2[i, :]
        jmin1, jmax1 = find_pix_borders(sp1, sought_value=0)
        jmin2, jmax2 = find_pix_borders(sp2, sought_value=0)
        image2d_merged[i, :] = sp1 + sp2
        useful1 = (jmin1 != -1) and (jmax1 != naxis1)
        useful2 = (jmin2 != -1) and (jmax2 != naxis1)
        if useful1 and useful2:
            jmineff = max(jmin1, jmin2)
            jmaxeff = min(jmax1, jmax2)
            image2d_merged[i, jmineff:(jmaxeff+1)] /= 2

    # return result
    image_merged = fits.PrimaryHDU(data=image2d_merged.astype(np.float32),
                                   header=image_header)

    return image_merged


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: merge 2 EMIR images averaging the common'
                    ' region'
    )

    # positional arguments
    parser.add_argument("infile1",
                        help="Input FITS file name #1",
                        type=argparse.FileType('rb'))
    parser.add_argument("infile2",
                        help="Input FITS file name #2",
                        type=argparse.FileType('rb'))
    parser.add_argument("outfile",
                        help="Output FITS file name",
                        type=lambda x: arg_file_is_new(parser, x, mode='wb'))

    # optional arguments
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting/debugging" +
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # read input FITS files
    hdulist1 = fits.open(args.infile1)
    hdulist2 = fits.open(args.infile2)

    image_merged = merge2images(
        hdulist1,
        hdulist2,
        debugplot=args.debugplot
    )

    # save result
    image_merged.writeto(args.outfile, overwrite=True)


if __name__ == "__main__":

    main()


#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Wavelength calibration parameters for each grism + filter configuration"""

from __future__ import division, print_function

import sys


def islitlet_progress(islitlet, islitlet_max):
    """Auxiliary function to print out progress in loop of slitlets.

    Parameters
    ----------
    islitlet : int
        Current slitlet number.
    islitlet_max : int
        Maximum slitlet number.

    """
    if islitlet % 10 == 0:
        cout = str(islitlet // 10)
    else:
        cout = '.'
    sys.stdout.write(cout)
    if islitlet == islitlet_max:
        sys.stdout.write('\n')
    sys.stdout.flush()


#
# Copyright 2013-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""AIV Recipes for EMIR"""


from numina.array.combine import median
from numina.core import Result, Requirement
from numina.processing.combine import basic_processing_with_combination

from emirdrp.core.recipe import EmirRecipe
import emirdrp.requirements as reqs
import emirdrp.products as prods


class TestSkyCorrectRecipe(EmirRecipe):

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = Requirement(prods.MasterIntensityFlat, 'Master Sky calibration')

    frame = Result(prods.ProcessedImage)

    def run(self, rinput):
        self.logger.info('starting simple sky reduction')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow, method=median)
        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        # Update SEC to 0
        hdr['SEC'] = 0
        
        result = self.create_result(frame=hdulist)

        return result

from numina.core import ObservationResult


class StareImageRecipeInputBuilder(object):
    '''Class to build StareImageRecipe inputs from the Observation Results.

       Fetches SKY calibration image from the archive


    '''

    def __init__(self, dal):
        self.dal = dal
        self.sky_image = None

    def buildRecipeInput(self, obsres):

        if self.sky_image is None:
            print('obtaining SKY image')
            sky_cal_result = self.dal.getLastRecipeResult("EMIR", "EMIR", "IMAGE_SKY")
            self.sky_image = sky_cal_result['elements']['skyframe']

        obsres['master_sky'] = self.sky_image
        newOR = ObservationResult()
        newOR.frames = obsres['frames']
        obsres['obresult'] = newOR
        newRI = StareImageRecipeInput(**obsres)

        return newRI


StareImageRecipeInput = TestSkyCorrectRecipe.RecipeInput
TestSkyCorrectRecipe.InputBuilder = StareImageRecipeInputBuilder



#
# Copyright 2013-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""AIV Recipes for EMIR"""

from __future__ import division

import logging
import math

import numpy
import scipy.interpolate as itpl
import scipy.optimize as opz
from scipy import ndimage

from astropy.modeling import models, fitting
import astropy.wcs
import photutils
from photutils import CircularAperture

from numina.array.recenter import centering_centroid
from numina.array.utils import image_box
from numina.array.fwhm import compute_fwhm_2d_spline
from numina.constants import FWHM_G
from numina.array.fwhm import compute_fwhm_2d_simple
from numina.array.utils import expand_region

from emirdrp.core import EMIR_PIXSCALE
from .procedures import compute_fwhm_enclosed_direct
from .procedures import compute_fwhm_enclosed_grow
from .procedures import moments
from .procedures import AnnulusBackgroundEstimator
from .procedures import image_box2d

_logger = logging.getLogger(__name__)


# returns y,x
def compute_fwhm(img, center):
    X = numpy.arange(img.shape[0])
    Y = numpy.arange(img.shape[1])

    bb = itpl.RectBivariateSpline(X, Y, img)
    # We assume that the peak is in the center...
    peak = bb(*center)[0, 0]

    def f1(x):
        return bb(x, center[1]) - 0.5 * peak

    def f2(y):
        return bb(center[0], y) - 0.5 * peak

    def compute_fwhm_1(U, V, fun, center):

        cp = int(math.floor(center + 0.5))

        # Min on the rigth
        r_idx = V[cp:].argmin()
        u_r = U[cp + r_idx]

        if V[cp + r_idx] > 0.5 * peak:
            # FIXME: we have a problem
            # brentq will raise anyway
            pass

        sol_r = opz.brentq(fun, center, u_r)

        # Min in the left
        rV = V[cp - 1::-1]
        rU = U[cp - 1::-1]
        l_idx = rV.argmin()
        u_l = rU[l_idx]
        if rV[l_idx] > 0.5 * peak:
            # FIXME: we have a problem
            # brentq will raise anyway
            pass

        sol_l = opz.brentq(fun, u_l, center)
        fwhm = sol_r - sol_l
        return fwhm

    U = X
    V = bb.ev(U, [center[1] for _ in U])
    fwhm_x = compute_fwhm_1(U, V, f1, center[0])

    U = Y
    V = bb.ev([center[0] for _ in U], U)

    fwhm_y = compute_fwhm_1(U, V, f2, center[1])

    return center[0], center[1], peak, fwhm_x, fwhm_y


# Background in an annulus, mode is HSM
def compute_fwhm_global(data, center, box):
    sl = image_box(center, data.shape, box)
    raster = data[sl]

    background = raster.min()
    braster = raster - background

    newc = center[0] - sl[0].start, center[1] - sl[1].start
    try:
        res = compute_fwhm(braster, newc)
        return (res[1] + sl[1].start, res[0] + sl[0].start,
                res[2], res[3], res[4])
    except ValueError as error:
        _logger.warning("%s", error)
        return center[1], center[0], -99.0, -99.0, -99.0
    except Exception as error:
        _logger.warning("%s", error)
        return center[1], center[0], -199.0, -199.0, -199.0


# returns x,y
def gauss_model(data, center_r):
    sl = image_box(center_r, data.shape, box=(4, 4))
    raster = data[sl]

    # background
    background = raster.min()

    b_raster = raster - background

    new_c = center_r[0] - sl[0].start, center_r[1] - sl[1].start

    yi, xi = numpy.indices(b_raster.shape)

    g = models.Gaussian2D(amplitude=b_raster.max(), x_mean=new_c[
                          1], y_mean=new_c[0], x_stddev=1.0, y_stddev=1.0)
    f1 = fitting.LevMarLSQFitter()  # @UndefinedVariable
    t = f1(g, xi, yi, b_raster)

    mm = t.x_mean.value + sl[1].start, t.y_mean.value + \
        sl[0].start, t.amplitude.value, t.x_stddev.value, t.y_stddev.value
    return mm

 
def recenter_char(data, centers_i, recenter_maxdist, recenter_nloop, recenter_half_box, do_recenter):
    
    # recentered values
    centers_r = numpy.empty_like(centers_i)
    # Ignore certain pinholes
    compute_mask = numpy.ones((centers_i.shape[0],), dtype='bool')
    status_array = numpy.ones((centers_i.shape[0],), dtype='int')
    
    for idx, (xi, yi) in enumerate(centers_i):
        # A failsafe
        _logger.info('for pinhole %i', idx)
        _logger.info('center is x=%7.2f y=%7.2f', xi, yi)
        if (xi > data.shape[1] - 5 or xi < 5 or
                yi > data.shape[0] - 5 or yi < 5):
            _logger.info('pinhole too near to the border')
            compute_mask[idx] = False
            centers_r[idx] = xi, yi
            status_array[idx] = 0
        else:
            if do_recenter and (recenter_maxdist > 0.0):
                kk = centering_centroid(
                    data, xi, yi,
                    box=recenter_half_box,
                    maxdist=recenter_maxdist, nloop=recenter_nloop
                )
                xc, yc, _back, status, msg = centering_centroid(
                    data, xi, yi,
                    box=recenter_half_box,
                    maxdist=recenter_maxdist, nloop=recenter_nloop
                )
                _logger.info('new center is x=%7.2f y=%7.2f', xc, yc)
                # Log in X,Y format
                _logger.debug('recenter message: %s', msg)
                centers_r[idx] = xc, yc
                status_array[idx] = status
            else:
                centers_r[idx] = xi, yi
                status_array[idx] = 0

    return centers_r, compute_mask, status_array


def pinhole_char(data, ncenters, box=4, recenter_pinhole=True, maxdist=10.0):

    ibox = (box, box)

    # convert FITS x, y coordinates (pixel center 1)
    # to python/scipy/astropy (pixel center 0 and x,y -> y,x)

    centers_i = ncenters[:, 0:2] - 1
    centers_r, cmask, starr = recenter_char(data, centers_i,
                                            recenter_maxdist=maxdist,
                                            recenter_nloop=10,
                                            recenter_half_box=ibox,
                                            do_recenter=recenter_pinhole
                                            )

    # Result 0
    nresults = 11
    mm0 = numpy.empty((centers_r.shape[0], nresults))
    mm0[:, 0:2] = centers_r
    mm0[:, 2] = starr
    mm0[:, 3:] = -99

    # compute the FWHM without fitting
    for idx, (xc, yc) in enumerate(centers_r):
        _logger.info('For pinhole %i', idx)
        if cmask[idx]:
            _logger.info('compute model-free FWHM')
            try:
                res1 = compute_fwhm_global(data, (yc, xc), box=ibox)
                fmt1 = 'x=%7.2f y=%7.2f peak=%6.3f fwhm_x=%6.3f fwhm_y=%6.3f'
                _logger.info(fmt1, *res1)
                mm0[idx, 3:6] = res1[2:]
            except Exception as error:
                _logger.exception("unable to obtain FWHM, %s", error)

            _logger.info('compute Gaussian 2Dfitting')
            try:
                res2 = gauss_model(data, (yc, xc))
                fmt2 = 'x=%7.2f y=%7.2f peak=%6.3f stdev_x=%6.3f stdev_y=%6.3f'
                _logger.info(fmt2, *res2)
                mm0[idx, 6:9] = res2[2:]
            except Exception as error:
                _logger.exception("unable to obtain FWHM, %s", error)
        else:
            _logger.info('skipping')

    # Photometry in coordinates
    # x=centers_r[:,0]
    # y=centers_r[:,1]
    # with radius 2.0 pixels and 4.0 pixels

    apertures = [2.0, 4.0]
    for idx, rad in enumerate(apertures):
        _logger.info('compute photometry with aperture radii %s', rad)
        apertures = CircularAperture(centers_r[:,:2], r=rad)
        faper = photutils.aperture_photometry(data, apertures)
        mm0[:, 9+idx] = faper['aperture_sum']
        _logger.info('done')

    # Convert coordinates to FITS
    mm0[:, 0:2] += 1
    return mm0


def pinhole_char2(
    data, ncenters,
    recenter_pinhole=True,
    recenter_half_box=5,
    recenter_nloop=10,
    recenter_maxdist=10.0,
    back_buff=3,
    back_width=5,
    phot_niter=10,
    phot_rad=8
):

    sigma0 = 1.0
    rad = 3 * sigma0 * FWHM_G
    box = recenter_half_box
    recenter_half_box = (box, box)

    # convert FITS x, y coordinates (pixel center 1)
    # to python/scipy/astropy (pixel center 0 and x,y -> y,x)

    centers_i = ncenters[:, 0:2] - 1

    centers_r, cmask, starr = recenter_char(data, centers_i, recenter_maxdist,
                                            recenter_nloop, recenter_half_box,
                                            do_recenter=recenter_pinhole
                                            )

    # Number of results
    nresults = 35
    mm0 = numpy.empty((centers_r.shape[0], nresults))
    mm0[:, 0:2] = centers_i
    mm0[:, 2:4] = centers_r
    mm0[:, 4] = starr
    mm0[:, 5:] = -99

    # Fitter
    fitter = fitting.LevMarLSQFitter()  # @UndefinedVariable
    rplot = phot_rad

    for idx, (x0, y0) in enumerate(centers_r):
        _logger.info('For pinhole %i', idx)
        if not cmask[idx]:
            _logger.info('skipping')
            # Fill result with -99
            continue

        # Initial photometric radius
        rad = 3.0
        # Loop to find better photometry radius and background annulus
        irad = rad
        bck = 0.0
        for i in range(phot_niter):
            phot_rad = rad
            # Sky background annulus
            rs1 = rad + back_buff
            rs2 = rs1 + back_width
            _logger.debug('Iter %d, annulus r1=%5.2f r2=%5.2f', i, rs1, rs2)
            bckestim = AnnulusBackgroundEstimator(r1=rs1, r2=rs2)

            # Crop the image to obtain the background
            sl_sky = image_box2d(x0, y0, data.shape, (rs2, rs2))
            raster_sky = data[sl_sky]
            # Logical coordinates
            xx0 = x0 - sl_sky[1].start
            yy0 = y0 - sl_sky[0].start
            # FIXME, perhaps we dont need to crop the image
            try:
                bck = bckestim(raster_sky, xx0, yy0)
                _logger.debug('Iter %d, background %f in '
                              'annulus r1=%5.2f r2=%5.2f',
                              i, bck, rs1, rs2)
            except Exception as error:
                _logger.warning('Error in background estimation %s', error)
                break

            # Radius of the fit
            fit_rad = max(rplot, rad)

            sl = image_box2d(x0, y0, data.shape, (fit_rad, fit_rad))
            part = data[sl]
            # Logical coordinates
            xx0 = x0 - sl[1].start
            yy0 = y0 - sl[0].start

            yy, xx = numpy.mgrid[sl]
            _logger.debug('Iter %d, radial fit', i)

            # Photometry
            dist = numpy.sqrt((xx - x0) ** 2 + (yy - y0) ** 2)
            phot_mask = dist < fit_rad
            r1 = dist[phot_mask]
            part_s = part - bck
            f1 = part_s[phot_mask]
            # Fit radial profile
            model = models.Gaussian1D(amplitude=f1.max(), mean=0, stddev=1.0)
            model.mean.fixed = True  # Mean is always 0.0

            g1d_f = fitter(model, r1, f1, weights=(r1 + 1e-12) ** -1)

            rpeak = g1d_f.amplitude.value
            # sometimes the fit is negative
            rsigma = abs(g1d_f.stddev.value)

            rfwhm = rsigma * FWHM_G

            rad = 2.5 * rfwhm
            _logger.debug('Iter %d, new rad is %f', i, rad)
            if abs(rad - irad) < 1e-3:
                # reached convergence
                _logger.debug('Convergence in iter %d', i)
                break
            else:
                irad = rad
        else:
            _logger.debug('no convergence in photometric radius determination')

        _logger.info('background %6.2f, r1 %7.2f r2 %7.2f', bck, rs1, rs2)
        mm0[idx, 5:5 + 3] = bck, rs1, rs2
        aper_rad = rad
        ca = CircularAperture([(xx0, yy0)], aper_rad)
        m = photutils.aperture_photometry(part_s, ca)
        flux_aper = m['aperture_sum'][0]
        _logger.info('aper rad %f, aper flux %f', aper_rad, flux_aper)
        mm0[idx, 8:8 + 2] = aper_rad, flux_aper

        _logger.info('Radial fit, peak: %f fwhm %f', rpeak, rfwhm)

        try:
            dpeak, dfwhm, smsg = compute_fwhm_enclosed_direct(
                part_s, xx0, yy0, maxrad=fit_rad)
            _logger.info('Enclosed direct, peak: %f fwhm %f', dpeak, dfwhm)
        except Exception as error:
            _logger.warning('Error in compute_fwhm_enclosed_direct %s', error)
            dpeak, dfwhm = -99.0, -99.0

        try:
            eamp, efwhm, epeak, emsg = compute_fwhm_enclosed_grow(
                part_s, xx0, yy0, maxrad=fit_rad)
            _logger.info('Enclosed fit, peak: %f fwhm %f', epeak, efwhm)
        except Exception as error:
            _logger.warning('Error in compute_fwhm_enclosed_grow %s', error)
            eamp, efwhm, epeak, emsg = [-99.0] * 4

        mm0[idx, 10:10 + 6] = epeak, efwhm, dpeak, dfwhm, rpeak, rfwhm

        try:
            res_simple = compute_fwhm_2d_simple(part_s, xx0, yy0)
            _logger.info('Simple, peak: %f fwhm x %f fwhm %f', *res_simple)
            mm0[idx, 16:16 + 3] = res_simple
        except Exception as error:
            _logger.warning('Error in compute_fwhm_2d_simple %s', error)
            mm0[idx, 16:16 + 3] = -99.0

        try:
            res_spline = compute_fwhm_2d_spline(part_s, xx0, yy0)
            _logger.info('Spline, peak: %f fwhm x %f fwhm %f', *res_spline)
            mm0[idx, 19:19 + 3] = res_spline
        except Exception as error:
            _logger.warning('Error in compute_fwhm_2d_spline %s', error)
            mm0[idx, 19:19 + 3] = -99.0

        # Bidimensional fit
        # Fit in a smaller box
        fit2d_rad = int(math.ceil(fit_rad))

        fit2d_half_box = (fit2d_rad, fit2d_rad)
        sl1 = image_box2d(x0, y0, data.shape, fit2d_half_box)

        part1 = data[sl1]
        yy1, xx1 = numpy.mgrid[sl1]

        g2d = models.Gaussian2D(amplitude=rpeak, x_mean=x0, y_mean=y0,
                                x_stddev=1.0, y_stddev=1.0)
        g2d_f = fitter(g2d, xx1, yy1, part1 - bck)

        res_gauss2d = (g2d_f.amplitude.value,
                       g2d_f.x_mean.value + 1,  # FITS coordinates
                       g2d_f.y_mean.value + 1,  # FITS coordinates
                       g2d_f.x_stddev.value * FWHM_G,
                       g2d_f.y_stddev.value * FWHM_G,
                       g2d_f.theta.value
                       )

        _logger.info('Gauss2d, %s', res_gauss2d)
        mm0[idx, 22:22 + 6] = res_gauss2d
        # Moments
        moments_half_box = fit2d_half_box
        res_moments = moments(data, x0, y0, moments_half_box)
        _logger.info('Mxx %f Myy %f Mxy %f e %f pa %f', *res_moments)

        mm0[idx, 28:28 + 5] = res_moments

    # Photometry in coordinates
    # x=centers_r[:,0]
    # y=centers_r[:,1]
    # with radius 2.0 pixels and 4.0 pixels

    apertures = [2.0, 4.0]
    for idx, rad in enumerate(apertures):
        _logger.info('compute photometry with aperture radii %s', rad)
        apertures = CircularAperture(centers_r[:,:2], r=rad)
        faper = photutils.aperture_photometry(data, apertures)
        mm0[:, 33+idx] = faper['aperture_sum']
        _logger.info('done')

    # FITS coordinates
    mm0[:, :4] += 1

    return mm0


def shape_of_slices(tup_of_s):
    return tuple(m.stop - m.start for m in tup_of_s)


def normalize(data):
    b = data.max()
    a = data.min()

    if b != a:
        data_22 = (2 * data - b - a) / (b - a)
    else:
        data_22 = data - b
    return data_22


def normalize_raw(arr):
    """Rescale float image between 0 and 1

    This is an extension of image_as_float of scikit-image
    when the original image was uint16 but later was
    processed and transformed to float32

    Parameters
    ----------
    arr; ndarray

    Returns
    -------
      A ndarray mapped between 0 and 1
    """

    # FIXME: use other limits acording to original arr.dtype
    # This applies only to uint16 images
    # As images were positive, the range is 0,1

    return numpy.clip(arr / 65535.0, 0.0, 1.0)


def char_slit(data, regions, box_increase=3, slit_size_ratio=4.0):

    result = []

    for r in regions:
        _logger.debug('initial region %s', r)
        oshape = shape_of_slices(r)

        ratio = oshape[0] / oshape[1]
        if (slit_size_ratio > 0)  and (ratio < slit_size_ratio):
            _logger.debug("this is not a slit, ratio=%f", ratio)
            continue

        _logger.debug('initial shape %s', oshape)
        _logger.debug('ratio %f', ratio)
        rp = expand_region(r, box_increase, box_increase,
                           start=0, stop=2048)
        _logger.debug('expanded region %r', rp)
        ref = rp[0].start, rp[1].start
        _logger.debug('reference point %r', ref)
    
        datas = data[rp]

        c = ndimage.center_of_mass(datas)
    
        fc = datas.shape[0] // 2
        cc = datas.shape[1] // 2
        _logger.debug("%d %d %d %d", fc, cc, c[0], c[1])

        _peak, fwhm_x, fwhm_y = compute_fwhm_2d_simple(datas, c[1], c[0])

        _logger.debug('x=%f y=%f', c[1] +  ref[1], c[0] +  ref[0])
        _logger.debug('fwhm_x %f fwhm_y %f', fwhm_x, fwhm_y)

        # colrow = ref[1] + cc + 1, ref[0] + fc + 1

        result.append([c[1] +  ref[1] + 1, c[0] +  ref[0] + 1, fwhm_x, fwhm_y])

    return result
